// $Header: /n/basmati/da/bedk/speech/src/frontend/RCS/Nonlinearity.cc,v 1.2 1998/05/29 18:20:47 bedk Exp bedk $
//
// $Log: Nonlinearity.cc,v $
// Revision 1.2  1998/05/29 18:20:47  bedk
// added sqrt compression
//
// Revision 1.1  1998/02/07 00:17:41  bedk
// Initial revision
//
//
// Simple memoryless nonlinearity

static const char rcs_id[] = "$Header: /n/basmati/da/bedk/speech/src/frontend/RCS/Nonlinearity.cc,v 1.2 1998/05/29 18:20:47 bedk Exp bedk $";

#include "Nonlinearity.h"

#include <assert.h>
#include <math.h>

const CMatrix
Nonlinearity::processBlock(const CMatrix in,
                           bool& lastTime)
{
  assert(in.m == nc);
  assert(in.n <= ioblocking);

  if (state == End)
    {
      lastTime = true;
      return CMatrix(nc,
                     0,
                     ioblocking,
                     out_buffer);
    }

  if (in.n == 0)
    {
      if (lastTime) { state = End; }
      return CMatrix(nc,
                     0,
                     ioblocking,
                     out_buffer);
    }

  if (state == Start) { state = Continuing; }

  const float *in_p = in.data;
  const float *in_endp = in.data + in.m * in.s;
  float* res_row_startp = out_buffer;

  if (method == HALFWAVE)
    {
      while (in_p != in_endp)
        {
          const float *in_row_p = in_p;
          const float *in_row_endp = in_p + in.n;
          float *res_row_p = res_row_startp;

          while (in_row_p != in_row_endp)
            {
              float t = *in_row_p++;
              if (t < 0.0) { t = 0.0; }
              *res_row_p++ = t;
            }

          in_p += in.s;
          res_row_startp += ioblocking;
        }
    }
  else if (method == SQRT)
    {
      while (in_p != in_endp)
        {
          const float *in_row_p = in_p;
          const float *in_row_endp = in_p + in.n;
          float *res_row_p = res_row_startp;

          while (in_row_p != in_row_endp)
            {
              float t = *in_row_p++;
              if (t < 0.0)
                {
                  *res_row_p++ = -sqrt(-t);
                }
              else
                {
                  *res_row_p++ = sqrt(t);
                }
            }

          in_p += in.s;
          res_row_startp += ioblocking;
        }
    }
  else if (method == POW)
    {
      while (in_p != in_endp)
        {
          const float *in_row_p = in_p;
          const float *in_row_endp = in_p + in.n;
          float *res_row_p = res_row_startp;

          while (in_row_p != in_row_endp)
            {
              float t = *in_row_p++;
              if (t < 0.0)
                {
                  *res_row_p++ = -pow(-t, exponent);
                }
              else
                {
                  *res_row_p++ = pow(t, exponent);
                }
            }

          in_p += in.s;
          res_row_startp += ioblocking;
        }
    }
  else if (method == SQR)
    {
      while (in_p != in_endp)
        {
          const float *in_row_p = in_p;
          const float *in_row_endp = in_p + in.n;
          float *res_row_p = res_row_startp;

          while (in_row_p != in_row_endp)
            {
              float t = *in_row_p++;
              *res_row_p++ = t * t;
            }

          in_p += in.s;
          res_row_startp += ioblocking;
        }
    }

  return CMatrix(nc,
                 in.n,
                 ioblocking,
                 out_buffer);
}
