// $Header$
//
// $Log$
//
// Object to do various forms of padding on a data stream
//

#ifndef Padding_h_INCLUDED
#define Padding_h_INCLUDED

#include <stdlib.h>

#include "Blocker.h"
#include "Vector.h"
#include "Matrix.h"

class Padding : public Blocker
{
public:
  enum pad_method
  {
    ZERO = 0,
    COPY
  };

  Padding(const size_t a_ioblocking,
          const size_t a_nc,
          pad_method a_method,
          const size_t a_nLeading,
          const size_t a_nTrailing)
    : ioblocking(a_ioblocking),
      nc(a_nc),
      method(a_method),
      nLeading(a_nLeading),
      nTrailing(a_nTrailing),
      points_in_buffer(0),
      prev_out_len(0),
      buffer(new float[nc * (ioblocking + nLeading + nTrailing + 1)])
  {};

  ~Padding() { delete buffer; };

  const CVector processBlock(const CVector in,
                             bool& lastTime);

  const CMatrix processBlock(const CMatrix in,
                             bool& lastTime);

private:
  const size_t ioblocking;      // max. length input we'll ever see

  const size_t nc;              // number of input channels
  
  const pad_method method;      // type of padding to do

  const size_t nLeading;        // number of leading points of padding
                                // to do

  const size_t nTrailing;       // number of trailing points of
                                // padding to do

  size_t points_in_buffer;      // number of valid points in buffer

  size_t prev_out_len;          // number of points output last time

  float* const buffer;          // output buffer
};

#endif // Padding_h_INCLUDED
