// $Header: /u/drspeech/src/bedk_frontend/RCS/modspec.cc,v 1.2.1.1 2000/08/03 19:42:36 dpwe Exp $
//
// $Log: modspec.cc,v $
// Revision 1.2.1.1  2000/08/03 19:42:36  dpwe
//  - added fake nyquist limitation option.  But I don't think it
//    quite works, so I'm still using resampling first.
//
// Revision 1.2  1998/06/24 18:30:50  bedk
// Fixed buffer sizing to handle downsampling by factors that don't evenly
// divide the ioblocking factor.
//
// Revision 1.1  1998/06/22 21:47:13  bedk
// Initial revision
//
//
// Modulation spectrogram front end for Broadcast News experiments

#include <string.h>
#include <math.h>

#include "OnlineSpeechInput.h"
#include "OnlfeatureOutput.h"
#include "PowspecFiltbank.h"
#include "Nonlinearity.h"
#include "Padding.h"
#include "SOSFilter.h"
#include "Trimming.h"
#include "Agc_Kohlrausch.h"

#include "arguments.h"

// general args
// 
size_t ioblocking=8000;
float sf=8000;

// filterbank params
// 
size_t nfft = 256;
size_t nwin = 256;
size_t nstep = 128;
float fbw = 1.0;
float fst = 1.0;
float ftr = 1.0;
float flpc = 200.0;
float fnyq_frac = 1.0;

char* filtshape = NULL;
class filtShape : public argsAction {
  PowspecFiltbank::FilterShape shape; 
public:
  filtShape() { shape = PowspecFiltbank::TRIANGULAR; }
  void act(char*&filtshape) { 
    if (!strncmp(filtshape,"TRAPEZOIDAL",strlen(filtshape))) {
      shape = PowspecFiltbank::TRAPEZOIDAL;
    }
    else if (!strncmp(filtshape,"TRIANGULAR",strlen(filtshape))) {
      shape = PowspecFiltbank::TRIANGULAR;
    }
  }
  PowspecFiltbank::FilterShape filtershape() { return shape; }
} fshape;

char* filtscale = NULL;
class filtScale : public argsAction {
  PowspecFiltbank::FilterScale scale; 
public:
  filtScale() { scale = PowspecFiltbank::BARK; }
  void act(char*&filtscale) { 
    if (!strncmp(filtscale,"CONST_Q",strlen(filtscale))) {
      scale = PowspecFiltbank::CONST_Q;
    }
    else if (!strncmp(filtscale,"BARK",strlen(filtscale))) {
      scale = PowspecFiltbank::BARK;
    }
  }
  PowspecFiltbank::FilterScale filterscale() { return scale; }
} fscale;

// Envelope filter params
//
char* efiltname = NULL;

// AGC params
//
float agctau1 = 0.0;
float agctau2 = 0.0;

ARGS ARGS::Args[] = {
  // general
  // 
  ARGS("ioblocking",ARGS::Opt,ioblocking,"The I/O blocking size (samples)"),
  ARGS("sf",ARGS::Opt,sf,"Sample Frequency (Hz)"),
  
  // FFT-based filterbank
  // 
  ARGS("nfft",ARGS::Opt,nfft,"# FFT points"),
  ARGS("nwin",ARGS::Opt,nwin,"# window points"),
  ARGS("nstep",ARGS::Opt,nstep,"# window step points"),
  ARGS("fbw",ARGS::Opt,fbw,"Filterbank filter bandwidth"),
  ARGS("fst",ARGS::Opt,fst,"Filterbank filter step"),
  ARGS("ftr",ARGS::Opt,ftr,"Filterbank filter transition region"),
  ARGS("flpc",ARGS::Opt,flpc,"Filterbank filter lowpass cutoff (Hz)"),
  ARGS("fshape",ARGS::Opt,filtshape,"Shape of filters in filterbank",fshape),
  ARGS("fscale",ARGS::Opt,filtscale,"Scale of filters in filterbank",fscale),

  // dpwe addition
  ARGS("fnyq",ARGS::Opt,fnyq_frac,"Fraction of available bandwidth to use"),

  // Envelope filter
  //
  ARGS("efilt",ARGS::Req,efiltname,"File holding envelope filter spec."),

  // AGC parameters
  //
  ARGS("agctau1",ARGS::Req,agctau1,"AGC time constant, in ms."),
  ARGS("agctau2",ARGS::Req,agctau2,"AGC time constant, in ms."),

  ARGS()
};

int
main(int argc, char** argv)
{
  ARGS::parse(argc,argv);

  OnlineSpeechInput osi((size_t) ioblocking, "-");

  // Generate a Hamming window
  // 
  Vector hamming(nwin);
  for (size_t i = 0; i < nwin; i++)
    hamming.data[i] = 0.54 - 0.46 * cos(2.0 * M_PI * i / (nwin - 1));

  const size_t nchan = 
    PowspecFiltbank::computeNumChannels(fscale.filterscale(),
                                        fbw,
                                        fst,
                                        sf,
                                        flpc, 
					fnyq_frac);

  OnlfeatureOutput ofo(nchan, "-");

  bool EOFile = false;

  // Loop over all of the input utterances
  //

  while (!EOFile)
    {
      bool EOUtt = false;
      bool EOProc = false;

      PowspecFiltbank fb((size_t) ioblocking,
                         nfft,
                         nstep,
                         hamming,
                         fbw,
                         fst,
                         ftr,
                         sf,
                         flpc,
                         fshape.filtershape(),
                         fscale.filterscale(), 
			 false, 
			 fnyq_frac);

      Nonlinearity nonl(ioblocking,
                        nchan,
                        Nonlinearity::SQRT);

      SOSFilter ef(1 + (ioblocking / nstep),
                   nchan,
                   efiltname);
      
      Padding pad(1 + (ioblocking / nstep),
                  nchan,
                  Padding::COPY,
                  0,
                  ef.group_delay());
      
      Trimming trim(1 + (ioblocking / nstep),
                    nchan,
                    ef.group_delay(),
                    0);
      
      Agc_Kohlrausch agc1(1 + (ioblocking / nstep),
                          sf / nstep,
                          nchan,
                          agctau1);
      
      Agc_Kohlrausch agc2(1 + (ioblocking / nstep),
                          sf / nstep,
                          nchan,
                          agctau2);
      
      // Compute amplitude envelope signals and write out as online
      // feature stream
      // 
      while (!EOProc)
        {
          const CVector b1 = osi.processBlock(EOUtt, EOProc, EOFile);
          const CMatrix b2 = fb.processBlock(b1,EOProc);
          const CMatrix b3 = nonl.processBlock(b2,EOProc);
          const CMatrix b4 = pad.processBlock(b3,EOProc);
          const CMatrix b5 = ef.processBlock(b4,EOProc);
          const CMatrix b6 = trim.processBlock(b5,EOProc);
          const CMatrix b7 = agc1.processBlock(b6,EOProc);
          const CMatrix b8 = agc2.processBlock(b7,EOProc);
          ofo.write(b8, EOProc);
        }
    }
}


