. $Header: /u/drspeech/src/dr_scripts/RCS/dr_align.man,v 1.5 2000/02/16 21:25:28 fosler Exp $
.de EX		\" Begin Example
.ie \\n(.$ .nr EX \\$1n
.el \{\
.	ie n .nr EX 0n
.	el .nr EX 0n
.\}
.in +\\n(EXu
.if n .sp 1
.if t .sp .5v
.nf
.CW
.ft CB
..
.de EE		\" End example
.in -\\n(EXu
.fi
.}f             \" restore basic text attributes
.if n .sp 1
.if t .sp .5v
..
.de MS
.ie t \&\f(CB\\$1\f(NR(\\$2)\\$3
.el \fI\\$1\fP(\\$2)\\$3
..
.TH DR_ALIGN 1 "$Date: 2000/02/16 21:25:28 $" ICSI "ICSI SPEECH SOFTWARE"
.SH NAME
dr_align \- generic interface to forced-alignment (Viterbi) routines
.SH SYNOPSIS
.B dr_align
\fBactivation_file\fR=<file>
\fBactivation_fmt\fR=<format>
\fButt_words_file\fR=<file>
\fButt_words_file_type\fR=<type>
\fButt_start\fR=<num>
\fButt_count\fR=<num>
\fBdictionary\fR=<file>
\fBphi\fR=<file>
\fBphone_models\fR=<file>
\fBpriors\fR=<file>
\fBctm_file\fR=<output_file>
\fBctm_type\fR=<type>
\fIadditional decoder-specific options\fR
.SH DESCRIPTION
.I dr_align
is, at present, merely a definition of the programmatic interface to 
the forced-alignment stage in speech recognizer training, where a best 
alignment is found between a set of posterior probabilities (the result of 
passing an utterance through an acoustic model) and the transcribed words 
to which they are believed to correspond, given a particular set of 
pronunciation models.  Alignment is used primarily as part of an embedded 
training loop (e.g. within
.MS dr_embed 1 )
where acoustic models are successively refined by relabelling the training 
data with each generation of the model, then retraining a new model.
.P
There are several different programs or procedures through which forced 
alignment might be performed, some of which are currently available, 
others not.  The intention in writing this man page is to define an 
interface to the alignment process that will permit the maximum flexibility 
in switchin between these mechanisms, particularly as future mechanisms 
(such as forced alignment within 
.MS noway 1 )
become available.
.P
This man page describes a command-line program to perform alignment 
stand-alone.  However, this is a trivial wrapper around a Perl library 
function to do the same procedure, which can be very easily incorporated 
into larger recognition scripts.
.P
We came up with four different possible implementations for 
.I dr_align
: using 
.MS y0 1 , 
using 
.MS noway 1 
(when this function becomes available), and two methods using 
.MS fsdg 1 :
one, the current Cambridge two-pass approach to get word alignments, then 
align phones within each word, and second to use a different program to 
construct a phone-grammar for the whole utterance, 
then to generate phone alignments with a single call to 
.MS fsgd 1 .
.SH OPTIONS
The following options are accepted (defaults in parens):
.TP
.B activation_file=\fIfile\fR ("-")
The source of the input probabilities.  Specifying "-" means that 
activations are read from standard input.
.TP
.B activation_format=\fIfile\fR ("rapbin")
The format of the input probability stream.  "rapbin" means the standard 
ICSI probability stream, starting with the number of probabilities per 
frame as a 32-bit int, followed by frames of that many 32-bit floats 
preceded by a 32-bit integer frame count, which is -1 at utterance ends.  
Other allowable values are "lna" for lna files (in which case the 
number of probabilities per frame is inferred from the priors or 
phone_models file).
.TP
.B utt_words_file=\fIfile\fR (required)
The file containing the target word transcriptions of each utterance, 
if the format specified by \f4utt_words_file_type\f1.
.TP
.B utt_words_file_type=\fItype\fR ("ref")
Either "ref", meaning that \f4utt_words_file\f1 contains one word
sequence per line, with no other information, "tagref", meaning that
\f4utt_words_file\f1 contains one word sequence per line, where the
first word is interpreted as an utterance id, or "stm", meaning that
the utterances are specified in the DARPA/NIST STM file format.
.P
.PD 0
.B utt_start=\fInum\fR (0)
.TP
.PD 1
.B utt_count=\fInum\fR (0)
The range of utterances to decode. The first 
\f4utt_start\f1 segments in both the activation stream and the 
reference transcripts are skipped (but see \f4utt_start_ref\f1 below), 
and a maximum 
of \f4utt_count\f1 segments are processed  before the program 
exits.  \f4utt_count\f1 smaller than 1 means continue until 
the end of the file.
.TP
.B utt_start_ref=\fInum\fR (0)
The skew between the first utterance in reference transcripts and the 
first segment in the activation stream.  This value is added to 
\f4utt_start\f1 to get the actual number of segments skipped from the 
reference file.  Thus, if the activation stream starts with utterance 
100 of a known training set, \f4utt_start_ref\f1 should be 100 
but \f4utt_start\f1 should be zero (meaning that no input activation 
segments are skipped).
.TP
.B dictionary=\fIfile\fR (required)
Pronunciation models for every word used, in 
.MS noway 1
format.
.P
.PD 0
.B phi=\fIfile\fR
.P
.B phone_models=\fIfile\fR
.TP
.PD 1
.B priors=\fIfile\fR
The individual phone states used in the dictionary's pronunciation 
specifications are defined either by a phone_model file, or by a phi (phone 
information) file, as defined for 
.MS noway 1 .
Phi files also include the prior probabilities of each label, used to 
convert the input posterior probabilities into the scaled likelihoods used 
in HMM models.  Since phone_model files do not include this information, 
when using this option it is necessary additionally to specify a separate 
priors file (an ascii file with one 
probability per line, exactly matching the number of probabilities per 
input frame).
.TP
.B ctm_file=\fIfile\fR (required)
The alignment data is written to this file.  The NIST-defined CTM format 
is ascii with five or six fields per line: utterance number, 
channel indicator,start time (in seconds), duration (seconds again) and 
the label symbol, possibly followed a confidence score.  The 
symbol part may be phones, words or both as controlled by the 
\f4ctm_type\f1 option.  It is legal to specify "-" to mean that CTM 
information should be written as a stream to standard output; however, 
this precludes post-processing fixup such as the padding of non-aligning 
sentences optionally performed by 
.MS dr_align_efsg 1 .
.TP
.B ctm_type=\fItoken\fR ("phone+word")
The CTM file definition only provides for a single output token for 
each defined region, yet there are two types of units that we may 
be aligning: the transcribed words, and the individual phonemes 
constituting the chosen pronunciation for that word.  A ctm_type 
of "phone" will emit just phone symbols, and "word" selects 
just the word alignments.  To get both symbols at the same time, 
the "phone+word" option acts like "phone" except for the 
first phoneme in each word, the emitted symbol is the concatenation of 
the phone and word symbols, separated by a "+", e.g.
.EX
1 A 1.470 0.130 w5+one -4.8611
1 A 1.600 0.060 ah4
1 A 1.660 0.150 n5
1 A 1.810 0.060 tcl3+two -4.0411
1 A 1.870 0.080 t4
1 A 1.950 0.170 uw9
.EE
Note that in this example, confidence scores are only emitted for 
the lines containing words, but I'm not sure how we'd do it really.  
Also, perhaps the words should be tagged onto the 
.I last
phone of their pronunciations rather than the first.
.TP
.B ctm_overwrite=\fIbool\fR ("false")
If "true", the program will silently overwrite any existing CTM file 
with the same name as specified for output.  Otherwise, it is an 
error for the file to exist when the program is run.
.TP
.B steptime=\fItime_ms\fR (10.0 ms)
The step between each analysis frame; this is needed to generate the 
output alignment times in seconds from the frame index which is the 
internal unit of time.
.SH EXAMPLES
To generate forced alignments of the NUMBERS95 probabilities previously 
calculated in n95tr-model1.lna:
.EX
dr_align activation_file=n95tr-model1.lna activation_fmt=lna \\
  priors=labels/n95tr+cv-icsi56.priors \\
  utt_words_file=trans/n95tr.wrdstrings \\
  dictionary=lm/n95g-prune1.dict phone_models=lm/n95g-prune1.phon \\
  ctm_file=n95-phn.ctm ctm_type=phone
.EE
From within Perl, 
.EX
@params={"activation_file=n95tr-model1.lna", \\
	 "activation_fmt=lna", \\
	 "priors=labels/n95tr+cv-icsi56.priors", \\
	 "utt_words_file=trans/n95tr.wrdstrings", \\
	 "dictionary=lm/n95g-prune1.dict", \\
	 "phone_models=lm/n95g-prune1.phon", \\
	 "ctm_file=n95-phn.ctm", 
	 "ctm_type=phone"};
&dr_align @params;
.EE
.SH NOTES
.SH BUGS
The script described in this man page doesn't yet exist.
.P
It's not clear how to get both word and phone alignments out, since 
the basic CTM format doesn't support this.  The existing 
ctm_type=phone+word is a hack, due to Eric.
.SH AUTHOR
Dan Ellis  <dpwe@ICSI.Berkeley.EDU>, from the ideas of Eric Fusslier, David Johnson, Dan Gildea and Brian Kingsbury.
.SH SEE ALSO
.MS noway 1 ,
.MS y0 1 ,
.MS fsgd 1 ,
.MS dr_embed 1 .
