. $Header: /u/drspeech/src/dr_scripts/RCS/dr_align_efsg.man,v 1.15 2002/03/12 23:22:52 gelbart Exp $
.de EX		\" Begin Example
.ie \\n(.$ .nr EX \\$1n
.el \{\
.	ie n .nr EX 0n
.	el .nr EX 0n
.\}
.in +\\n(EXu
.if n .sp 1
.if t .sp .5v
.nf
.CW
.ft CB
..
.de EE		\" End example
.in -\\n(EXu
.fi
.}f             \" restore basic text attributes
.if n .sp 1
.if t .sp .5v
..
.de MS
.ie t \&\f(CB\\$1\f(NR(\\$2)\\$3
.el \fI\\$1\fP(\\$2)\\$3
..
.TH DR_ALIGN_EFSG 1 "$Date: 2002/03/12 23:22:52 $" ICSI "ICSI SPEECH SOFTWARE"
.SH NAME
dr_align_efsg \- an implementation of forced-alignment using finite-state grammars
.SH SYNOPSIS
.B dr_align_efsg
\fIoptions\fR
.SH DESCRIPTION
.B dr_align_efsg
is an implementation of the generic forced-alignment interface described in 
.MS dr_align 1 .
It is based around 
.MS gen_efsg 1 
which converts word sequences into finite-state phone-grammars specific 
to each utterance, and 
.MS efsgd 1 
which decodes a probability stream derived from an utterance's features 
within the grammar derived from the word sequence; this is what forced 
alignment means.  The inputs, therefore, are an "activation stream" 
(typically an LNA file or stream from 
.MS qnsfwd 1 ), 
a word sequence for each utterance (as a CTM file or a simple text file), 
some way to generate pronunciations for each word (such as a 
.MS noway 1 
-format dictionary file) and the definitions of the individual phone 
Markov models (e.g. as a 
.MS noway 1 
phi file).  The output is a CTM file or stream containing the best 
alignment of the best pronunciation sequence for each utterance in turn.  
All this is as described in the man page for 
.MS dr_align 1 .
.P
In addition, 
.B dr_align_efsg
has the ability to synthesize pronunciations for out-of-vocabulary 
words missing from the dictionary, or word fragments marked with 
dashes in the transcription, by using a text-to-speech algorithm 
based on decision trees.  This greatly improves the robustness of 
generating alignments to large unconstrained training sets.
.SH OPTIONS
.B dr_align_efsg 
accepts all the options defined in the man page for 
.MS dr_align 1 .
The following extra options are also accepted, specific to 
this implementation. (Defaults are in parens):
.TP
.B valid_utts=\fIfile\fR ("valid_utts.$$")
It is possible that certain segments may fail to align, implying that 
the decoder was unable to find any reasonable path matching the 
words to the probabilities.  To account for this possibility, 
the program may write a list of the utterance numbers that were successfully 
aligned.  Moreover, if all segments aligned correctly, this can contain 
simply one word, "all".  Otherwise it is a list of utterance numbers, 
starting from zero, one per line.
.TP
.B verbose=\fInum\fR
The logging level.  Setting verbose=2 is a useful level for diagnosing
problems.  It will print out all of the things in the pipeline.  You can see 
then which command is actually breaking.  The level can go higher than 2.
.TP
.PD 0
.B ftr_pfile=\fIfile\fR ("")
.TP
.PD 1
.B zeropad=\fItime\fR (40.0 ms)
If a segment does not align, the finite state grammar decoder does not 
generate any output, but 
.B dr_align_efsg
will insert a dummy segment, consisting entirely of a dummy symbol, 
to permit a somewhat graceful recovery from this situation.  Since 
the final goal of forced-alignment is often a label pfile whose frame 
count should exactly match a training pfile, it is important for this 
dummy segment to match the exact length that would be generated by 
a good alignment.  In order to do this, you can point to the source 
pfile, as well as specifying the amount of padding that has 
been added at each end to account for the MLP context window, and 
then dummy filler segments will be inserted for non-aligning utterances.  
Currently, it is acceptable not to specify this pfile, but in that case 
it is an error if any segments fail to align.
.TP
.B efsgd_prog=\fIpath\fR ("efsgd")
The name or path of the binary to execute as the extended finite-state 
grammar decoder.  This is the Softsound 
.MS fsgd 1
as modified locally by Eric.
.TP
.B efsgd_opt_args=\fIstring\fR ("")
Optional additional arguments to pass in the invocation of \f4efsgd\f1.
.SS Options for \f4gen_efsg(1)\f1:
.TP
.B treelist=\fIfile\fR ("")
Name of file containing list of text-to-speech trees that 
.MS gen_efsg 1 
can use to create pronunciations for unknown words.  
These are generated by
.MS eric_fosler_lussier 1 .
.TP
.PD 0
.B silprob=\fIprob\fR (0.001)
.TP
.PD 0
.B silwrd=\fIsym\fR ("<SIL>")
.TP
.PD 1
.B silphn=\fIsym\fR ("h#")
For inserting silences between words, these define the unigram probability 
to associate with such an interword silence, and the phone and 
word (for phn+word output format) symbols to use when it is inserted.
.TP 
.B fragment_samplect=\fInum\fR (10)
The number of randomly-selected words used by 
.MS gen_efsg 1 
to determine pronunciations for partial words.
.TP 
.B random_seed=\fInum\fR (0)
The random seed used when generating pronunciations for partial words.
This is specified to ensure that successive runs with the same 
data will produce identical results.
.TP
.B prunelevel=\fIprob\fR (0.01)
Letter-to-phoneme rules with a probability below this threshold are 
pruned from the pronunciation generation procedure.
.TP
.B alphabet=\fIstring\fR ("A-Za-z\\-\\'\\.\\#...")
Set of letters for which letter-to-phoneme rules have been defined i.e. 
the characters treated as significant when deriving TTS pronunciations 
from spellings.

.SH EXAMPLES
To generate a phone-only CTM file containing the alignments of a
particular BN show (represented by its LNA file), using a non-installed 
version of the \f4efsgd\f1 binary, including dummy
segments where the alignment failed, and using the default installed
TTS rules for generating any words not in the dictionary:
.EX
dr_align_efsg \\
  efsgd_prog=/u/drspeech/opt/softsound/v_1.2.19_hacked/src/Sprog/efsgd \\
      activation_file=$SHOW.lna \\
    activation_format=lna \\
       utt_words_file=$SHOW.stm \\
  utt_words_file_type=stm \\
            ftr_pfile=$SHOW.pf \\
              zeropad=64.0 \\
             steptime=16.0 \\
         phone_models=/u/drspeech/data/bn/models/CI54_BN97.model \\
               priors=/u/drspeech/data/bn/models/CI54_BN97.priors \\
           dictionary=/u/drspeech/data/bn/lex/bn96-65k.dct \\
             treelist=/u/drspeech/data/gen_efsg_tts/default.treelist \\
           valid_utts=$SHOW.validutts \\
             ctm_file=$SHOW.ctm \\
             ctm_type=phone \\
        ctm_overwrite=true
.EE
To generate a realigned label file for an entire BN training set, 
skipping the first $SKIPFIRST segments in the source pfile, 
using the full-coverage bn98 master dictionary to avoid the need 
for any TTS trees (note that 
recovery from non-aligning sentences is not possible in this mode 
where the CTM is written to the standard output):
.EX
STEPTIME=16.0
PADTIME=64.0
BNROOT=$SPEECH_DIR/data/bn
qnsfwd \\
    fwd_sent_start=$SKIPFIRST \\
    ftr1_file=$PFILE \\
    ftr1_norm_file=$NORMS \\
    ftr1_ftr_count=$NFTRS \\
    ftr1_window_offset=0 \\
    ftr1_window_len=$CW \\
    window_extent=$CW \\
    init_weight_file=$WTSFILE \\
    mlp3_input_size=$IUS \\
    mlp3_hidden_size=$HUS \\
    mlp3_output_size=$OUS \\
    mlp3_output_type=softmax \\
    log_file=$CWD/qnsfwd-$TAG.log \\
    activation_format=lna \\
    activation_file=- \\
| dr_align_efsg \\
    activation_format=lna \\
    utt_words_file=$STMFILE \\
    utt_words_file_type=stm \\
    utt_start_ref=$SKIPFIRST \\
    steptime=$STEPTIME \\
    phi=$BNROOT/models/CI54_BN98.align3.phi \\
    dictionary=$BNROOT/lex/bn98-master.dct \\
    ctm_file=- \\
| labels2pfile \\
    ctmfile=- \\
    steptime=$STEPTIME \\
    windowtime=$STEPTIME \\
    zeropad=$PADTIME \\
    phoneset=$SPEECH_DIR/data/phonesets/cam54.phset \\
    samplerate=16000 \\
    force=1 \\
    pfile=$PFLABFILE
.EE
.SH NOTES
As currently implemented, 
.B dr_align_efsg
is a tiny perl program that loads the module \f4dr_align_efsg.pm\f1 
and passes all its arguments on to it.  That module runs 
.MS feacat 1 
to normalize the input features, then invokes the \f4gen_efsg.pm\f1 
perl module for grammar definitions, setting up a couple of pipes to 
feed the \f4efsgd\f1 process.  It then tracks the stderr output of 
that process to trap non-aligning segments, which it substitutes 
with dummy segments if possible in a post-processing stage.
.P
There must be the same number of utterances in the transcription
file (\f4utt_words_file\f1) as in the activation file (\f4activation_file\f1).
They should be in the same order in both files.
.SH BUGS
The post-processing to pad-out non-aligning utterances is a big pain 
and it would be much cleaner if this information were somehow generated 
by \f4efsgd\f1 (which would then remove the need for ftr_pfile and zeropad).
.P
If the TTS is needed (meaning that there are unknown words resulting
in very bushy grammars) \f4efsgd\f1 tends to use up all available
memory for the longest utterances in the Broadcast News training set
(i.e. those of about 30,000 frames or more, correspoding to about 8
minutes).  It would be nice to trap these also and treat them as
non-aligning.  For now, the solution is to use the full-coverage 
training dictionary.
.P
It seems that a \f4phi\f1 file must be used rather than 
specifying \f4phone_models\f1 and \f4priors\f1 files?
.P 
This program can try to put in bad alignment markers for sentences
that didn't align.  This way, you have in your label file a marker for
those sentences, so that you don't have to build a new pfile/label
file for the sentences that align.  However Eric, writing
in January 2002, is not sure he got around to implementing this
for the ctm_type=phone only (instead of ctm_type=phone+word) case.
.P
When fsgd (which efsgd is based on) was written, they only had the
concept that you would be aligning word strings, not phone strings.
This could cause deceptive error messages.  For example, the error
message \f4efsgd: word not found in dictionary: ???\f1 could mean that
the phone (not word) ??? was found in the pronunciation dictionary but
was not in the phoneset.
.P
Utterances with an empty transcription (due to no words being spoken)
will confuse the aligner.  Instead of leaving transcriptions blank for
such utterances, use a silence word.  Or leave those utterances out.
.SH AUTHOR
Eric Fosler-Lussier <fosler@ICSI.Berkeley.EDU> (man page by Dan Ellis <dpwe@icsi.berkeley.edu>, with a few additions by David Gelbart <gelbart@icsi.berkeley.edu>).
.SH SEE ALSO
.MS dr_align 1 ,
.MS gen_efsg 1 ,
.MS efsgd 1 .
