/*
    $Header: /u/spert/src/fltvec/php/RCS/muladd_vffvf_vf.c,v 1.4 1995/10/13 00:15:08 davidj Exp $

    PHIPAC code for a muladd_vffvf_vf (SAXPY).
*/

#include <fltv-config.h>
#include <stddef.h>
#include <fltv-defs.h>
#include <fltv-scansrc.h>

#ifdef DECL_php_muladd_vffvf_vf
void php_muladd_vffvf_vf(const size_t n, const float* xvec,
			 const float s, const float *yvec,
			 float* res);
#endif
#ifdef DECL_php_muladd_vddvd_vd
void php_muladd_vddvd_vd(const size_t n, const double* xvec,
			 const double s, const double *yvec,
			 double* res);
#endif

#ifdef DEFN_php_muladd_vffvf_vf
#define FP_TYPE float
#define ROUT_NAME php_muladd_vffvf_vf
#endif

#ifdef DEFN_php_muladd_vddvd_vd
#define FP_TYPE double
#define ROUT_NAME php_muladd_vddvd_vd
#endif

void
ROUT_NAME(const size_t n,
          const FP_TYPE* xvec, const FP_TYPE s, const FP_TYPE *yvec,
          FP_TYPE* res)
{
    const FP_TYPE *const x_end_b8p = xvec + (n & ~7);
    const FP_TYPE *const x_end_b2p = xvec + (n & ~1);

    while (xvec != x_end_b8p)
    {
        const FP_TYPE sum0 = s * xvec[0] + yvec[0];
        const FP_TYPE sum1 = s * xvec[1] + yvec[1];
        const FP_TYPE sum2 = s * xvec[2] + yvec[2];
        const FP_TYPE sum3 = s * xvec[3] + yvec[3];
        const FP_TYPE sum4 = s * xvec[4] + yvec[4];
        const FP_TYPE sum5 = s * xvec[5] + yvec[5];
        const FP_TYPE sum6 = s * xvec[6] + yvec[6];
        const FP_TYPE sum7 = s * xvec[7] + yvec[7];

        xvec += 8; yvec += 8;

        res[0] =  sum0;
        res[1] =  sum1;
        res[2] =  sum2;
        res[3] =  sum3;
        res[4] =  sum4;
        res[5] =  sum5;
        res[6] =  sum6;
        res[7] =  sum7;

        res += 8;
    }

    /* Do remainder in pairs. */
    while (xvec != x_end_b2p)
    {
        const FP_TYPE sum0 = s * xvec[0] + yvec[0];
        const FP_TYPE sum1 = s * xvec[1] + yvec[1];

        xvec += 2; yvec += 2;

        res[0] =  sum0;
        res[1] =  sum1;

        res += 2;
    }

    /* Do last one. */
    if (n & 1)
    {
        *res = s * (*xvec) + (*yvec);
    }
}

