/*
** $Header: /u/spert/src/fltvec/tor/RCS/tor_div_gfgf_gf.S,v 1.5 1995/07/08 00:53:44 davidj Exp $
*/

#include "tor/tordef.h"
#include "m_regdef.h"

	.globl tor_div_gfgf_gf
#ifdef DECL_tor_div_gfgf_gf 
void tor_div_gfgf_gf(void); 
#endif

#define vA va0			/* Input A				*/
#define vB va1			/* Input B				*/
#define vZ vv2			/* Output Z				*/

#define vZSign vt0		/* Sign of result			*/
#define vEitherIsSubnormal vt1	/* Flag for either vA or vB is		*/
				/*     subnormal			*/
#define vAbsA vt4		/* vA shifted up one bit to remove sign	*/
#define vAbsB vt5		/* vB shifted up one bit to remove sign	*/
#define vAbsA2 vv2		/* vA shifted up one bit to remove sign	*/
#define vAExpOffset vt3		/* Fix-up to vA exponent to account for	*/
				/*     subnormal normalization		*/
#define vBExpOffset vv2		/* Fix-up to vB exponent to account for	*/
				/*     subnormal normalization		*/
#define savedRA t9		/* Return address			*/
#define vAExp2 vt1
#define vExpTooLarge vt2
#define vBExp2 vt1
#define vRecipBEstimate0 vtt	/* First estimate of reciprocal of	*/
				/*     significand of vB		*/
#define vBHigh vt1		/* vB significand shifted down to only	*/
				/*     16 bits				*/
#define vRecipBEstimate vtt	/* Estimated reciprocal of significand	*/
				/*     of vB				*/
#define vRecipBEstimate1 va0	/* Second estimate of reciprocal of	*/
				/*     significand of vB		*/
#define vRemainder vv2		/* Running remainder			*/
#define vRemainderHigh vv3	/* vRemainder shifted down to only	*/
				/*     16 bits				*/
#define vPartialZSignif va0	/* Partial significand quotient		*/
#define vMultipliedBackHigh va2	/* High part of vB significand times	*/
				/*     partial quotient			*/
#define vMultipliedBackLow va3	/* Low part of vB significand times	*/
				/*     partial quotient			*/
#define vRawZSignif vt2		/* Raw significand of result		*/
#define vBLow vt3		/* Low 8 bits of vB significand		*/
#define vBSignif vv3		/* vB significand			*/
#define vZSignifAdjust vt3	/* Significand quotient adjustment	*/
#define vAdjustedBExp vtt	/* Adjusted exponent of vB		*/
#define vExpAdjust vv3		/* Normalization adjustment for result	*/
				/*     exponent				*/
#define vAdjustedAExp vt1	/* Adjusted exponent of vA		*/
#define vZSignif vv2		/* Rounded result significand		*/
#define vZExp vt1		/* Under-biased exponent of result	*/
#define vBothAreFiniteAndBIsNonzero vt3
#define vInvalidDivision vtt	/* Flag for invalid operation		*/
#define vAIsInfiniteOrBIsZero vt3
#define vACode vt4
#define vBCode vt5
#define vDivisionByZero vt3
#define vNaNResult vtt
#define vCount16 vt1		/* vCountLeadingZerosA & 0x10		*/
#define vCount8 vt2		/* vCountLeadingZerosA & 0x08		*/
#define vLowIndex va0		/* Index into low zeros-count table	*/
#define vHighIndex va1		/* Index into high zeros-count table	*/
#define vCountLeadingZerosA vt3	/* Number of leading zeros of vAbsA	*/
#define vShiftedA va2		/* vAbsA shifted for count of leading	*/
				/*     zeros				*/

#define temp0 v0
#define temp1 v1
#define temp2 t4
#define config0 t5
#define config1 t6
#define config2 t7
#define config3 t8
#define config4 t9

	.ent tor_div_gfgf_gf

/*
-------------------------------------------------------------------------------
Routine:  tor_div_gfgf_gf

Input:
	vlr
	va0	vA
	va1	vB

Output:
	vv2	vZ

Not preserved:
	vtt
	vv3
	va0--va3
	vt0--vt5
	v0--v1
	t4--t9

Function:

The single-precision floating-point vectors vA and vB are element-wise divided
and the single-precision result vector returned in vZ.  The vlr register
controls the elements of vZ actually updated.  This operation conforms to the
IEEE Standard for Binary Floating-point Arithmetic, except as follows:
(1) Only nearest/even rounding is supported; (2) No inexact flags are raised;
and (3) No signalling NaNs are recognized.  All NaNs are quiet.  For all cases
in which the result of the subtraction is a NaN, the default MIPS NaN is
returned.  The bit patterns of input NaNs are not preserved.

Exceptions for any of the element-wise divisions are signalled by raising
element-wise flags in the global variables tor_fltv_flags_invalid32,
tor_fltv_flags_divbyzero32, tor_fltv_flags_overflow32, and tor_fltv_flags_underflow32.  No traps
are taken for any exceptions.  Underflow is never signalled for results that
are not subnormal.

-------------------------------------------------------------------------------
*/

tor_div_gfgf_gf:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	va0	vA
|	va1	vB
|	vt0	vZSign
|	vt1	vEitherIsSubnormal
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Find the sign of the result from the exclusive-or of the signs of the inputs.
Shift up vA and vB each by one bit to eliminate their signs.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	addu.vv vAbsA,vA,vA
	addu.vv vAbsB,vB,vB
	li config1,(FX_SHLZA|FX_A_XOR_B|FX_CLIP_SLTZ)
	fxadd.vv vZSign,vA,vB,config1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set vEitherIsSubnormal nonzero if vA is subnormal and nonzero, or if vB is
subnormal or zero.  (Ideally, vEitherIsSubnormal would be zero if vB is zero,
but this way is cheaper.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp1,(1<<24)
	sltu.vs vEitherIsSubnormal,vAbsB,temp1
	li config2,(FX_BU|FX_LURES|FX_CLIP_PASS|FX_GTZ)
	fxsub.sv vEitherIsSubnormal,temp1,vAbsA,config2

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set-up a constant and a pipeline configuration to be used in extracting the
significand of vB.  (This occurs here only for instruction scheduling
purposes.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,0xFF000000
	li config0,(FX_SHLZA|(25*FX_SHLAM)|FX_A_ANDNOT_B\
	            |FX_SEPSHAM|(9*FX_SHRAM)|FX_SHRNR|FX_CLIP_PASS)

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vA or vB is subnormal and vA is nonzero, then a subnormal fix-up is needed.
If a subnormal fix-up is not needed for any vector element, skip the code to
perform the fix-up.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	ctvu zero,vcond
	fnez.v vEitherIsSubnormal
	cfvu temp2,vcond
	beqz temp2,inputsNormalized

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vAbsA2,vBExpOffset
|	vt0	vZSign
|	vt3	vAExpOffset
|	vt4	vAbsA
|	vt5	vAbsB
|	t9	savedRA
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vA is subnormal, normalize it by shifting up.  Set vAExpOffset to the number
of bits shifted (zero if normal, infinity, or NaN).  Do the same for vB,
setting vBExpOffset to the number of bits shifted.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	move savedRA,ra
	mov.vv vAbsA2,vAbsA
	mov.vv vAbsA,vAbsB
	jal normalizeA
	mov.vv vAbsB,vAbsA
	mov.vv vAbsA,vAbsA2
	mov.vv vBExpOffset,vAExpOffset
	jal normalizeA
	move ra,savedRA

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vBExpOffset
|	vt0	vZSign
|	vt1	vAExp2,vBExp2
|	vt2	vExpTooLarge
|	vt3	vAExpOffset
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vA is zero, make sure it stays zero by clearing vBExpOffset.  (See below.)
Do the same for vB and vAExpOffset.  Hence for zero inputs, a subnormal
normalization of the other input will not be reflected in the zero input.  But
of course, zero inputs give degenerate results anyway.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	cmveqz.vv vBExpOffset,vAbsA,vzero
	cmveqz.vv vAExpOffset,vAbsB,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vA is not too large, add vBExpOffset to the exponent of vAbsA, so that the
quotient of the modified vAbsA and the normalized vAbsB will be the same as
that of the original inputs.  If vAbsA is a NaN or infinity, or if the exponent
of vAbsA is too large to be added to, leave it alone.  (If vAbsA is too large,
dividing by the normalized vAbsB is bound to overflow anyway; the exponent
adjustment does not matter.)
    Do the same for vAExpOffset and vAbsB.  If vAbsB is a NaN or infinity, or
if the exponent of vAbsB is too large to be added to, leave it alone.  (If
vAbsB is too large, dividing it into the normalized vAbsB is bound to underflow
to zero.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp1,24
	li temp2,0xFE
	srlv.vs vAExp2,vAbsA,temp1
	addu.vv vAExp2,vBExpOffset
	sltu.sv vExpTooLarge,temp2,vAExp2
	cmvnez.vv vBExpOffset,vExpTooLarge,vzero
	sllv.vs vBExpOffset,vBExpOffset,temp1
	addu.vv vAbsA,vAbsA,vBExpOffset
	srlv.vs vBExp2,vAbsB,temp1
	addu.vv vBExp2,vAExpOffset
	sltu.sv vExpTooLarge,temp2,vBExp2
	cmvnez.vv vAExpOffset,vExpTooLarge,vzero
	sllv.vs vAExpOffset,vAExpOffset,temp1
	addu.vv vAbsB,vAbsB,vAExpOffset

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set-up a constant and a pipeline configuration to be used in extracting the
significand of vB.  (This occurs here only for instruction scheduling
purposes.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,0xFF000000
	li config0,(FX_SHLZA|(25*FX_SHLAM)|FX_A_ANDNOT_B\
	            |FX_SEPSHAM|(9*FX_SHRAM)|FX_SHRNR|FX_CLIP_PASS)

inputsNormalized:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vRecipBEstimate0,vRecipBEstimate
|	va0	vRecipBEstimate1
|	vt0	vZSign
|	vt1	vBHigh
|	vt3	vBLow
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Separate out the high 16 bits of the divisor significand from the low 8 bits,
and make the leading significand bit explicit.
    Zero, infinity, and NaN divisors are (mis)treated as though they were
normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	fxadd.vs vBHigh,vAbsB,temp0,config0
	li temp2,7
	sllv.vs vBLow,vAbsB,temp2

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set vRecipBEstimate0 to an approximation of the reciprocal of vB, accurate to
within 1/23.  vRecipBEstimate0 is a fixed point number less than 1, with binary
point between bits 15 and 16.  (The top 16 bits are all zero.)
    (vRecipBEstimate0 is calculated as 0x17505/0x10000 - vB/2, which is
approximately 1.4571075 - vB/2.  The result is truncated to 16 bits.  The
leading 1 in 0x17505 is not actually needed in the calculation because the
result is known to be less than 1, so only the fractional part is relevant.)
    Zero, infinity, and NaN divisors continue to be (mis)treated as though they
were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp1,0x7505
	subu.sv vRecipBEstimate0,temp1,vBHigh

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Given vRecipBEstimate0, calculate in vRecipBEstimate a better approximation to
the reciprocal of vB, accurate to within 1/16384.  vRecipBEstimate is a fixed-
point number less than 1, with binary point between bits 15 and 16.  (The top
16 bits are all zero.)
    (First an intermediate reciprocal estimate, accurate to within 1/263, is
formed in vRecipBEstimate1 with the expression

	( 2 - vB*vRecipBEstimate0 )*vRecipBEstimate0.

This is a standard formula for cancelling some of the inaccuracy of a
reciprocal approximation to obtain a more accurate approximation.  Since
multiplication can only be done directly with 16-bit operands, only the first
16 bits of vB are used.  The inner product is truncated to a fixed-point number
with binary point between bits 14 and 15.  Subtraction from 2 can be performed
by negation because the result is known to be less than 2, so only the bits
below 2 are relevant.  vRecipBEstimate1 is a fixed-point number less than 1
with binary point between bits 15 and 16.
    (vRecipBEstimate is then formed in the same way from vRecipBEstimate1.
This time the limit on accuracy is rounding, so the estimate cannot be improved
by further iterations.)
    Zero, infinity, and NaN divisors continue to be (mis)treated as though they
were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config2,(FX_AU|FX_BU|FX_SHRL|(16*FX_SHLAM)|FX_RDOWN|FX_CLIP_PASS)
	subu config3,config2,(1*FX_SHLAM)
	fxmul.vv vRecipBEstimate1,vRecipBEstimate0,vBHigh,config2
	negu.vv vRecipBEstimate1,vRecipBEstimate1
	fxmul.vv vRecipBEstimate1,vRecipBEstimate0,vRecipBEstimate1,config3
	fxmul.vv vRecipBEstimate,vRecipBEstimate1,vBHigh,config2
	negu.vv vRecipBEstimate,vRecipBEstimate
	fxmul.vv vRecipBEstimate,vRecipBEstimate1,vRecipBEstimate,config3

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vRemainder
|	vv3	vRemainderHigh
|	vtt	vRecipBEstimate
|	va0	vPartialZSignif
|	va2	vMultipliedBackHigh
|	va3	vMultipliedBackLow
|	vt0	vZSign
|	vt1	vBHigh
|	vt2	vRawZSignif
|	vt3	vBLow
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Make the leading significand bit of the dividend vA explicit and shift up to
extract the dividend significand, leaving one remaining exponent bit.  Start
with the initial remainder being the entire dividend significand.  Because the
previous shift is one bit shy, bit 31 of vRemainder is trash--the binary point
occurs between bits 29 and 30.  (Bit 31 is always a 1, but that is irrelevant.
When the first partial product is multiplied back and subtracted out of
vRemainder, this extra bit will be discarded.)  Set vRemainderHigh to the top
16 bits of the initial remainder.  (That is, bits 15 through 30 of vRemainder.
Bit 31 of vRemainder is not included in vRemainderHigh.)
    Zero, infinity, and NaN dividends are (mis)treated as though they were
normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	or.vs vRemainder,vAbsA,temp0
	li temp1,6
	sllv.vs vRemainder,vRemainder,temp1
	fxadd.vs vRemainderHigh,vAbsA,temp0,config0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set up some pipeline configurations used multiple times.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config0,(FX_AU|FX_BU|FX_CLIP_PASS)
	li config3,((11*FX_SHLAM)|FX_SEPSHAM|(16*FX_SHRAM)|FX_RDOWN)
	addu config4,config0,((11*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Determine a partial quotient vPartialZ by multiplying the most significant bits
of vRemainder (same as vA) by vRecipBEstimate, and then truncating to a 16-bit
fixed-point number with binary point between bits 14 and 15.  The 16-bit value
vPartialZSignif approximates the desired significand quotient with an absolute
error less than 1/6694.  Tentatively set vRawZSignif to vPartialZSignif,
shifting up by 16 bits to place the binary point between bits 30 and 31.
    Multiply vPartialZSignif back into vB, and subtract the product from
vRemainder.  The remainder is signed, since vPartialZSignif could be more or
less than the correct quotient.  The absolute value of the remainder must be
less than vB/6694, which is less than 1/3347.  Shift up the remainder by
11 bits, leaving vRemainder as a signed fixed-point number with binary point
between implicit bits 40 and 41.  Set vRemainderHigh to the top 16 bits of
vRemainder.
    (vPartialZSignif is purposefully truncated to 16 bits to make calculation
of an exact remainder cheaper.)
    Zero, infinity, and NaN inputs are (mis)treated as though they were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	addu config1,config0,((16*FX_SHLAM)|FX_RDOWN)
	fxmul.vv vPartialZSignif,vRemainderHigh,vRecipBEstimate,config1
	fxmul.vv vMultipliedBackHigh,vPartialZSignif,vBHigh,config0
	subu.vv vRemainder,vRemainder,vMultipliedBackHigh
	addu config2,config0,((5*FX_SHLAM)|FX_RDOWN)
	fxmul.vv vMultipliedBackLow,vPartialZSignif,vBLow,config2
	fxsub.vv vRemainderHigh,vRemainder,vMultipliedBackLow,config3
	fxsub.vv vRemainder,vRemainder,vMultipliedBackLow,config4
	li temp0,16
	sllv.vs vRawZSignif,vPartialZSignif,temp0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set up a pipeline configuration used multiple times.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config0,(FX_BU|FX_CLIP_PASS)

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Determine a second partial quotient vPartialZSignif by multiplying
vRemainderHigh by vRecipBEstimate and truncating to a fixed-point number with
binary point between bits 25 and 26.  vPartialZSignif approximates
vRemainder/(significand of vB) with an absolute error less than 1/13735000.
Add vPartialZSignif to vZSignif so that vZSignif approximates the desired
significand quotient with an absolute error less than 1/13735000.
    Multiply the new partial quotient back into vB, and subtract the product
from vRemainder.  The absolute value of the remainder must be less than
vB/13735000, which is less than 1/6867000.  Shift up the remainder again by
11 bits, leaving vRemainder as a signed fixed-point number with binary point
between implicit bits 51 and 52.  Set vRemainderHigh to the top 16 bits of
vRemainder.
    (Again, vPartialZSignif is purposefully truncated to 16 bits to make
calculation of an exact remainder cheaper.)
    Zero, infinity, and NaN inputs continue to be (mis)treated as though they
were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	addu config1,config0,((15*FX_SHLAM)|FX_RDOWN)
	fxmul.vv vPartialZSignif,vRemainderHigh,vRecipBEstimate,config1
	fxmul.vv vMultipliedBackHigh,vPartialZSignif,vBHigh,config0
	subu.vv vRemainder,vRemainder,vMultipliedBackHigh
	addu config2,config0,((5*FX_SHLAM)|FX_RDOWN)
	fxmul.vv vMultipliedBackLow,vPartialZSignif,vBLow,config2
	fxsub.vv vRemainderHigh,vRemainder,vMultipliedBackLow,config3
	fxsub.vv vRemainder,vRemainder,vMultipliedBackLow,config4
	li temp1,5
	sllv.vs vPartialZSignif,vPartialZSignif,temp1
	addu.vv vRawZSignif,vRawZSignif,vPartialZSignif

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Determine a third and final partial quotient vPartialZSignif by multiplying
vRemainderHigh by vRecipBEstimate and truncating to a fixed-point number with
binary point between bits 30 and 31.  vPartialZSignif approximates
vRemainder/(significand of vB) with an absolute error less than 1/(2.040*10^9).
Add vPartialZSignif to vZSignif so that vZSignif approximates the desired
significand quotient with an absolute error less than 1/(2.040*10^9), or almost
32 bits.
    Multiply the new partial quotient back into vB, and subtract the product
from vRemainder.  The absolute value of the remainder must be less than
vB/(2.040*10^9), which is less than 1/(1.020*10^9).  Shift up the remainder by
2 bits, leaving vRemainder as a signed fixed-point number with binary point
between implicit bits 53 and 54.  vRemainder must be between -vB/(43.65*10^9)
and vB/(2.040*10^9).
    Zero, infinity, and NaN inputs continue to be (mis)treated as though they
were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	addu config1,config0,((21*FX_SHLAM)|FX_RDOWN)
	fxmul.vv vPartialZSignif,vRemainderHigh,vRecipBEstimate,config1
	addu config2,config0,((8*FX_SHLAM)|FX_RDOWN)
	fxmul.vv vMultipliedBackLow,vPartialZSignif,vBLow,config2
	addu config3,config0,((2*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxsub.vv vRemainder,vRemainder,vMultipliedBackLow,config3
	fxmul.vv vMultipliedBackHigh,vPartialZSignif,vBHigh,config0
	li temp2,8
	sllv.vs vMultipliedBackHigh,vMultipliedBackHigh,temp2
	subu.vv vRemainder,vRemainder,vMultipliedBackHigh
	addu.vv vRawZSignif,vRawZSignif,vPartialZSignif

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vRemainder
|	vv3	vBSignif
|	vt0	vZSign
|	vt2	vRawZSignif
|	vt3	vZSignifAdjust
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Extract the significand of vB and make the leading bit explicit.
    Zero, infinity, and NaN inputs continue to be (mis)treated as though they
were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config4,(FX_SHLZA|(25*FX_SHLAM)|FX_A_ANDNOT_B\
	            |FX_SHRL|FX_SEPSHAM|(1*FX_SHRAM)|FX_SHRNR|FX_CLIP_PASS)
	li temp0,0xFF000000
	fxadd.vs vBSignif,vAbsB,temp0,config4

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vRemainder >= vB/(2^31), add 1/(2^31) to vZSignif and subtract vB/(2^31)
from vRemainder; else if vRemainder < 0, subtract 1/(2^31) from vZSignif.
    (It isn't necessary to add vB/(2^31) to vRemainder when vRemainder < 0,
because the remainder would still be nonzero; that is all the information that
is needed.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config3,(FX_NOT_B|FX_CLIP_SLTZ)
	fxadd.vv vZSignifAdjust,vBSignif,vRemainder,config3
	li temp1,-1
	cmvltz.vs vZSignifAdjust,vRemainder,temp1
	addu.vv vRawZSignif,vRawZSignif,vZSignifAdjust
	li config0,(FX_CLIP_PASS|FX_GEZ)
	fxsub.vv vRemainder,vRemainder,vBSignif,config0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Jam vRemainder into last bit of vZ.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	snez.vv vRemainder,vRemainder
	or.vv vRawZSignif,vRawZSignif,vRemainder

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vAdjustedBExp
|	vv3	vExpAdjust
|	vt0	vZSign
|	vt1	vAdjustedAExp,vZExp
|	vt2	vRawZSignif
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Calcuate the "under-biased" exponent of the result, by subtracting the
exponents of vA and vB and also taking into account whether vRawZSignif needs
to be shifted up by one bit to be normalized.  (The exponent is "under-biased"
because the integral portion of the significand will be added to it when the
result is packed.  In most cases, a 1 will be added to the exponent during
packing.  However, if a carry occurred during rounding of the significand, a
value of 2 will be added to the exponent during packing.)
    Zero, infinity, and NaN inputs continue to be (mis)treated as though they
were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config1,(FX_SHLZA|FX_A_AND_B|FX_BU\
	            |FX_SHRL|FX_SEPSHAM|(7*FX_SHRAM)|FX_NOROUND|FX_CLIP_PASS)
	li temp2,0x80000000
	fxadd.vs vExpAdjust,vRawZSignif,temp2,config1
	li config2,(FX_AU|FX_SEPSHAM|(24*FX_SHRAM)|FX_NOROUND|FX_CLIP_PASS)
	fxadd.vv vAdjustedAExp,vAbsA,vExpAdjust,config2
	li temp0,0x83000000
	fxadd.vs vAdjustedBExp,vAbsB,temp0,config2
	subu.vv vZExp,vAdjustedAExp,vAdjustedBExp

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vA is zero, force the result exponent to zero.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	cmveqz.vv vZExp,vAbsA,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vZSignif
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Normalize vRawZSignif.  The significand quotient will be close to normalized
anyway, but it can be off by one bit.  If vA is zero, force vRawZSignif to
zero.
    Zero divisors and infinity and NaN inputs continue to be (mis)treated as
though they were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config3,(FX_AU|FX_BU|FX_CLIP_PASS|FX_GEZ)
	fxadd.vv vRawZSignif,vRawZSignif,vRawZSignif,config3
	cmveqz.vv vRawZSignif,vAbsA,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Form the tentative result significand by shifting the raw result significand
down, rounding to nearest/even.
    Zero divisors and infinity and NaN inputs continue to be (mis)treated as
though they were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config4,(FX_BU|FX_SHRL|(8*FX_SHLAM)|FX_RNEARESTEVEN|FX_CLIP_PASS)
	fxadd.vv vZSignif,vzero,vRawZSignif,config4

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt3	vBothAreFiniteAndBIsNonzero
|	vt4	vAbsA
|	vt5	vAbsB
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vZExp is less than 0, the result is a subnormal or has underflowed.  If
vZExp is 0xFD or more, the result might have overflowed.  If any vector element
may have underflowed or overflowed, or if any divisor was zero, or if any input
was a NaN or infinity, jump to a special case handler.
    (Although the result has not actually overflowed in most cases in which
vZExp equals 0xFD, overflow does occur when vZExp is 0xFD and vRawZSignif >=
0xFFFFFF80.)
    (An exact zero result can occur only if vA was zero, in which case vZExp
was forced to zero above.  In all other cases, if vZExp is zero, the biased
exponent of the result will be at least 1.)
    (Note that temp1 is still live at the entrance to the special cases code.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp1,(0xFF<<24)
	sltu.vs vBothAreFiniteAndBIsNonzero,vAbsA,temp1
	li config0,(FX_AU|FX_BU|FX_CLIP_SLTZ|FX_GEZ)
	fxsub.vs vBothAreFiniteAndBIsNonzero,vAbsB,temp1,config0
	cmveqz.vv vBothAreFiniteAndBIsNonzero,vAbsB,vzero
	feqz.v vBothAreFiniteAndBIsNonzero
	cfvu temp1,vcond
	li temp2,0xFC
	fltu.sv temp2,vZExp
	cfvu temp2,vcond
	or temp2,temp1
	bnez temp2,specialCases

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vZSignif,vZ
|	vt0	vZSign
|	vt1	vZExp
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Add the integral portion of the significand to the under-biased exponent, and
pack the sign, biased exponent, and significand of the result.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config0,((23*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxadd.vv vZ,vZExp,vZSignif,config0
	li config1,((31*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxadd.vv vZ,vZSign,vZ,config1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Return.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j ra

specialCases:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If all inputs are finite and all divisors are nonzero, the only possible
special cases involve overflow or underflow of the result.  In that case, skip
the code to handle zero divisors and NaN and infinity inputs.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	beqz temp1,resultSpecialCasesOnly

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vInvalidDivision,vNaNResult
|	vv2	vZSignif
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt3	vAIsInfiniteOrBIsZero,vDivisionByZero
|	vt4	vAbsA,vACode
|	vt5	vAbsB,vBCode
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,(0xFF<<24)

	seq.vs vAIsInfiniteOrBIsZero,vAbsA,temp0
	cmveqz.vs vAIsInfiniteOrBIsZero,vAbsB,temp0

	seq.vv vInvalidDivision,vAbsA,vAbsB
	cmveqz.vv vInvalidDivision,vAIsInfiniteOrBIsZero,vzero

	ctvu zero,vcond
	fnez.v vInvalidDivision
	cfvu temp1,vcond
	lw temp2,tor_fltv_flags_invalid32
	or temp2,temp1
	sw temp2,tor_fltv_flags_invalid32

	li config0,(FX_AU|FX_BU|FX_SEPSHAM|(31*FX_SHRAM)|FX_RODD)
	fxsub.sv vACode,temp0,vAbsA,config0
	fxsub.sv vBCode,temp0,vAbsB,config0

	cmveqz.vv vZSignif,vBCode,vzero
	cmveqz.vv vZExp,vBCode,vzero
	li temp0,0x7F800000
	cmvnez.vs vZSignif,vAIsInfiniteOrBIsZero,temp0
	cmvnez.vv vZExp,vAIsInfiniteOrBIsZero,vzero
	cmvlez.vv vDivisionByZero,vACode,vzero
	cmvnez.vv vDivisionByZero,vInvalidDivision,vzero

	fnez.v vDivisionByZero
	cfvu temp1,vcond
	lw temp2,tor_fltv_flags_divbyzero32
	or temp2,temp1
	sw temp2,tor_fltv_flags_divbyzero32

	cmvltz.vv vNaNResult,vACode,vACode
	cmvltz.vv vNaNResult,vBCode,vBCode
	li temp0,0x7FBFFFFF
	cmvnez.vs vZSignif,vNaNResult,temp0
	cmvnez.vv vZExp,vNaNResult,vzero
	cmvnez.vv vZSign,vNaNResult,vzero

resultSpecialCasesOnly:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv2	vZSignif
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Continue with the special case result handler common to all single-precision
floating-point operations.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j tor_fltv_singlerange

normalizeA:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	va0	vLowIndex
|	va1	vHighIndex
|	va2	vShiftedA
|	vt1	vCount16
|	vt2	vCount8
|	vt3	vCountLeadingZerosA
|	vt4	vAbsA
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set vCountLeadingZerosALeadingZerosA to the number of leading zeros of vAbsA.
    (NOTE:  This block of code could all be replaced by a single count-leading-
zeros instruction.  See countlz_gi32_gi32.S for more refined commentary.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,(1<<16)
	sltu.vs vCount16,vAbsA,temp0
	li temp1,4
	sllv.vs vCount16,vCount16,temp1
	sllv.vv vShiftedA,vAbsA,vCount16
	li temp2,(1<<24)
	sltu.vs vCount8,vShiftedA,temp2
	li temp0,3
	sllv.vs vCount8,vCount8,temp0
	sllv.vv vShiftedA,vShiftedA,vCount8
	addu.vv vCount8,vCount16,vCount8
	li config0,(FX_SHLZA|FX_A_AND_B|(23*FX_SHLAM)|FX_NOROUND|FX_CLIP_PASS)
	li config1,(FX_SHLV|(28*FX_SHLAM)|FX_NOROUND|FX_CLIP_PASS|FX_GEZ)
	li temp1,(7<<25)
	fxadd.vs vLowIndex,vShiftedA,temp1,config0
	li temp2,0x76554444
	fxadd.sv vCountLeadingZerosA,temp2,vLowIndex,config1
	li temp0,(7<<28)
	addu config2,config0,(3*FX_SHLAM)
	fxadd.vs vHighIndex,vShiftedA,temp0,config2
	li temp1,0xF3221111
	fxadd.sv vCountLeadingZerosA,temp1,vHighIndex,config1
	cmvltz.vv vCountLeadingZerosA,vShiftedA,vzero
	addu.vv vCountLeadingZerosA,vCount8,vCountLeadingZerosA

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vt3	vCountLeadingZerosA,vAExpOffset
|	vt4	vAbsA
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vAbsA is subnormal, normalize it by shifting up.  Set vAExpOffset to the
number of bits shifted (zero if normal, infinity, or NaN).
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp2,7
	subu.vs vAExpOffset,vCountLeadingZerosA,temp2
	cmvltz.vv vAExpOffset,vAExpOffset,vzero
	sllv.vv vAbsA,vAbsA,vAExpOffset

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Return.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j ra

	.end tor_div_gfgf_gf

