/*
   $Header: /u/spert/src/intvec/testsuite/RCS/mat_mi16_mi16_test.c,v 1.2 1996/10/23 00:50:26 davidj Exp $

   Test matrix routines of the form op_vi16_vi16.
*/

#include <stddef.h>
#include <stdlib.h>

#include "intvec_test.h"


typedef void (*rtn_t)(size_t in_height, size_t in_width,
		      size_t in_stride, size_t res_stride,
		      const intv_int16_t* in, intv_int16_t* res);

/* Performs one test given input values and expected output. */

void
onetest_normal(rtn_t rtn, size_t in_height, size_t in_width,
	       size_t in_stride, size_t res_stride,
	       const intv_int16_t* in, const intv_int16_t* correct)
{
    const size_t res_width = in_height;
    const size_t res_height = in_width;

    /* Create padded copies of inputs. */
    intv_int16_t* c_in = rtst_padmat_clone_mi16(in_height, in_width,
						in_stride, in_stride, in);
    intv_int16_t* c_res = rtst_padmat_new_mi16(res_height, res_width,
					       res_stride);

    /* Run routine. */
    rtn(in_height, in_width, in_stride, res_stride, c_in, c_res);

    /* Check result. */
    rtst_checkeq_mi16mi16(res_height, res_width, res_stride, res_stride,
			  c_res, correct);

    /* Check input not corrupted. */
    rtst_checkeq_mi16mi16(in_height, in_width, in_stride, in_stride,
			  in, c_in);

    /* Free and check padding. */
    rtst_padmat_del_mi16(c_in);
    rtst_padmat_del_mi16(c_res);
}

void
trans_3x2(rtn_t rtn)
{
    const intv_int16_t in[] = { 0, -1, 1, -12345, 0x7fff, 0x8000};
    const intv_int16_t correct[] = { 0, 1, 0x7fff, -1, -12345, 0x8000 };
    
    onetest_normal(rtn, 3, 2, 2, 3, in, correct);
}

void
trans_3x2s4_s5(rtn_t rtn)
{
    const intv_int16_t in[] = { 0, -1, 0, 0,
				1, -12345, 0, 0,
				0x7fff, 0x8000, 0, 0};
    const intv_int16_t correct[] = { 0, 1, 0x7fff, 0, 0,
				     -1, -12345, 0x8000, 0, 0 };
    
    onetest_normal(rtn, 3, 2, 4, 5, in, correct);
}

/* Compare the results of two vector routines */

#ifdef NEVER
/* Vector version - needs modifying for matrix */
void
random_test(rtn_t nv_rtn, rtn_t test_rtn)
{
    const int numtests = rtst_numtests;
    const size_t maxlen = rtst_sizetests;
    intv_int16_t *in = rtst_padvec_new_vi16(maxlen);
    intv_int16_t *res = rtst_padvec_new_vi16(maxlen);

    int i;
    for (i=0; i<numtests; i++)
    {
        size_t len = (size_t) rtst_urand_i32i32_i32(1, maxlen);

        rtst_badrand_vi16(len, in);
        nv_rtn(len, in, res);

        rtst_log("len=%d\n", len);

        onetest_normal(test_rtn, len, in, res);
    }

    rtst_padvec_del_vi16(in);
    rtst_padvec_del_vi16(res);
}
#endif

void
test_trans(void)
{
    rtst_start("trans_mi16_mi16");
    rtst_log("testing nv.\n");
    trans_3x2(nv_trans_mi16_mi16);
    trans_3x2s4_s5(nv_trans_mi16_mi16);
/*
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    trans_6(tor_trans_mi16_mi16);
    random_test(nv_trans_mi16_mi16, tor_trans_mi16_mi16);
#endif
*/
    rtst_passed();
}

int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_trans();

    rtst_exit();

    return 0;
}
