/*
   $Header: /u/spert/src/intvec/testsuite/RCS/vecele_i32vi32_vi32_test.c,v 1.2 1996/01/03 21:14:28 davidj Exp $

   Test element wise vector routines of the form op_vfvf_vf.
*/

#include <stddef.h>
#include <stdlib.h>

#include "intvec_test.h"


typedef void (*rtn_t)(size_t len, intv_int32_t a, const intv_int32_t* b,
		      intv_int32_t* res);

/* Performs one test given input values and expected output. */

void
onetest_normal(rtn_t rtn, size_t len,
	       intv_int32_t a, const intv_int32_t* b,
	       const intv_int32_t* correct)
{
    /* Create padded copies of inputs. */
    intv_int32_t* c_b = rtst_padvec_clone_vi32(len, b);
    intv_int32_t* c_res = rtst_padvec_new_vi32(len);

    /* Run routine. */
    rtn(len, a, c_b, c_res);

    /* Check result. */
    rtst_checkeq_vi32vi32(len, c_res, correct);

    /* Check input not corrupted. */
    rtst_checkeq_vi32vi32(len, b, c_b);

    /* Free and check padding. */
    rtst_padvec_del_vi32(c_b);
    rtst_padvec_del_vi32(c_res);
}

void
sub1(rtn_t rtn)
{
    intv_int32_t a = 3;
    const intv_int32_t b[] = { -0, -1, 1, -123, 456, -0x7fffffff };
    const intv_int32_t correct[] = { 3, 4, 2, 126, -453, 0x80000002 };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
}

void
satsub1(rtn_t rtn)
{
    intv_int32_t a = 3;
    const intv_int32_t b[] = { -0, -1, 1, -123, 456, -0x7fffffff };
    const intv_int32_t correct[] = { 3, 4, 2, 126, -453, 0x7fffffff };
    
    onetest_normal(rtn, sizeof(a)/sizeof(intv_int32_t), a, b, correct);
}

/* Compare the results of two vector routines */

void
random_test(rtn_t nv_rtn, rtn_t test_rtn)
{
    const int numtests = rtst_numtests;
    const size_t maxlen = rtst_sizetests;
    intv_int32_t a;
    intv_int32_t *b = rtst_padvec_new_vi32(maxlen);
    intv_int32_t *res = rtst_padvec_new_vi32(maxlen);

    int i;
    for (i=0; i<numtests; i++)
    {
        size_t len = (size_t) rtst_urand_i32i32_i32(1, maxlen);

        a = rtst_badrand_i32();
        rtst_badrand_vi32(len, b);
        nv_rtn(len, a, b, res);

        rtst_log("len=%d\n", len);

        onetest_normal(test_rtn, len, a, b, res);
    }

    rtst_padvec_del_vi32(b);
    rtst_padvec_del_vi32(res);
}

void
test_sub(void)
{
    rtst_start("sub_i32vi32_vi32");
    rtst_log("testing nv.\n");
    sub1(nv_sub_i32vi32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    sub1(tor_sub_i32vi32_vi32);
    random_test(nv_sub_i32vi32_vi32, tor_sub_i32vi32_vi32);
#endif
    rtst_passed();
}

void
test_satsub(void)
{
    rtst_start("satsub_i32vi32_vi32");
    rtst_log("testing nv.\n");
    satsub1(nv_satsub_i32vi32_vi32);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    satsub1(tor_satsub_i32vi32_vi32);
    random_test(nv_satsub_i32vi32_vi32, tor_satsub_i32vi32_vi32);
#endif
    rtst_passed();
}

int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_sub();
    test_satsub();

    rtst_exit();

    return 0;
}
