/*
** $Id: binop_vi32i32_vi32.S,v 1.2 1996/01/03 20:56:02 davidj Exp $
**
** A general 32 bit vector/scalar binary operation routine, plus
** some instantiations
** davidj - Sat Oct 15 18:50:22 1994
*/

#include "tor/tordef.h"

/****************************************************************/

#ifdef DECL_tor_add_vi32i32_vi32
void tor_add_vi32i32_vi32(size_t len, const intv_int32_t* a, intv_int32_t b,
			  intv_int32_t* res);
#endif

#ifdef DEFN_tor_add_vi32i32_vi32
#define RTNNAME tor_add_vi32i32_vi32
#define OP_VS addu.vs
#endif

/****************************************************************/

#ifdef DECL_tor_sub_vi32i32_vi32
void tor_sub_vi32i32_vi32(size_t len, const intv_int32_t* a, intv_int32_t b,
			  intv_int32_t* res);
#endif

#ifdef DEFN_tor_sub_vi32i32_vi32
#define RTNNAME tor_sub_vi32i32_vi32
#define OP_VS subu.vs
#endif

/****************************************************************/

#ifdef DECL_tor_satadd_vi32i32_vi32
void tor_satadd_vi32i32_vi32(size_t len, const intv_int32_t* a, intv_int32_t b,
			     intv_int32_t* res);
#endif

#ifdef DEFN_tor_satadd_vi32i32_vi32
#define RTNNAME tor_satadd_vi32i32_vi32
#define OP_VS sadd.vs
#endif

/****************************************************************/

#ifdef DECL_tor_satsub_vi32i32_vi32
void tor_satsub_vi32i32_vi32(size_t len, const intv_int32_t* a, intv_int32_t b,
			     intv_int32_t* res);
#endif

#ifdef DEFN_tor_satsub_vi32i32_vi32
#define RTNNAME tor_satsub_vi32i32_vi32
#define OP_VS ssub.vs
#endif

/****************************************************************/

/*
** This routine requires 2 macros to be defined
**
** RTNNAME - the name this routine will be known by
** OP_VS - the binary vector-scalar op to perform
*/ 

#ifndef RTNNAME
	.err
#endif
#ifndef OP_VS
	.err
#endif

	.globl	RTNNAME


#define	a_v		vv1
#define res_v		vv2

#define len_r		a0
#define aptr_r		a1
#define b_r		a2
#define resptr_r	a3

#define fourvlr_r	t0
#define done_r		t1


	.ent	RTNNAME
RTNNAME:
 # Check the length - jump forward if > 32

        .set    noreorder
        bgtu    len_r, MAXVLEN, morethan32
        ctvu    len_r, vlr              # Assume len <= 32
        .set    reorder

 # The short case (len<=32) follows directly for better cache performance

        lw.v	a_v, aptr_r
	OP_VS	res_v, a_v, b_r
	sw.v	res_v, resptr_r
	j	ra
	
 # The len>32 case

morethan32:
	li	v0, MAXVLEN
	ctvu	v0, vlr
	lw.v	a_v, aptr_r
	li	fourvlr_r, MAXVLEN*4
	add	aptr_r, fourvlr_r
	OP_VS	res_v, a_v, b_r

loop:
	lwai.v	a_v, aptr_r, fourvlr_r
	swai.v	res_v, resptr_r, fourvlr_r
	subu	len_r, MAXVLEN
 	sltu	done_r, len_r, MAXVLEN+1
	OP_VS	res_v, a_v, b_r
	beqz	done_r, loop

 # Wind up after the loop

	ctvu	len_r, vlr
	sw.v	res_v, resptr_r
	j	ra

	.end	RTNNAME

