/*
** $Id: mul_vi32vi32_vi32.S,v 1.1 1995/08/31 00:28:46 davidj Exp $
**
** davidj - Wed Feb  1 11:41:05 1995
*/

#include <tor/tordef.h>

#ifdef DECL_tor_mul_vi32vi32_vi32
void tor_mul_vi32vi32_vi32(size_t len, const intv_int32_t* a,
			   const intv_int32_t *b, intv_int32_t* res);
#endif

#ifdef DEFN_tor_mul_vi32vi32_vi32

	.globl	tor_mul_vi32vi32_vi32

#define	a_v		vv1
#define b_v		vv2
#define ah_v		vv3	
#define bh_v		vv4
#define albl_v		vv5
#define ahbl_v		vv6
#define albh_v		vv7
#define res_v		vv8
#define temp_v		vv9

#define len_r		a0
#define aptr_r		a1
#define bptr_r		a2
#define resptr_r	a3

#define shift_r		t1
#define fourvlr_r	t2
#define conf_muluu_r	t3
#define conf_mulsu_r	t4
#define conf_mulus_r	t5
#define conf_shladd_r	t6
#define done_r		t7


	.ent	tor_mul_vi32vi32_vi32
tor_mul_vi32vi32_vi32:
 # Check the length - jump forward if > 32

        .set    noreorder
        bgtu    len_r, MAXVLEN, morethan32
        ctvu    len_r, vlr              # Assume len <= 32
        .set    reorder

 # The short case (len<=32) follows directly for better cache performance

        lw.v	a_v, aptr_r
	li	shift_r, 16
	li	conf_shladd_r, ((16*FX_SHLAM) | FX_SEPSHAM \
			       | FX_NOROUND | FX_CLIP_PASS)
	srav.vs	ah_v, a_v, shift_r
	lw.v	b_v, bptr_r
	li	conf_muluu_r, (FX_AU | FX_BU | FX_CLIP_PASS)
	fxmul.vv res_v, a_v, b_v, conf_muluu_r
	srav.vs	bh_v, b_v, shift_r
	li	conf_mulus_r, (FX_AU | FX_CLIP_PASS)
	fxmul.vv albh_v, a_v, bh_v, conf_mulus_r
	li	conf_mulsu_r, (FX_BU | FX_CLIP_PASS)
	fxadd.vv res_v, albh_v, res_v, conf_shladd_r
	fxmul.vv ahbl_v, ah_v, b_v, conf_mulsu_r
	fxadd.vv res_v, ahbl_v, res_v, conf_shladd_r
	sw.v	res_v, resptr_r
	j	ra	

 # The len>32 case

morethan32:
	li	v0, MAXVLEN
	ctvu	v0, vlr
	lw.v	a_v, aptr_r
	add	aptr_r, MAXVLEN*4
	li	fourvlr_r, MAXVLEN*4
	li	shift_r, 16
	srav.vs	ah_v, a_v, shift_r
	lwai.v	b_v, bptr_r, fourvlr_r
	li	conf_shladd_r, ((16*FX_SHLAM) | FX_SEPSHAM \
			       | FX_NOROUND | FX_CLIP_PASS)
	li	conf_muluu_r, (FX_AU | FX_BU | FX_CLIP_PASS)
	fxmul.vv albl_v, a_v, b_v, conf_muluu_r
	srav.vs	bh_v, b_v, shift_r
	li	conf_mulus_r, (FX_AU | FX_CLIP_PASS)
	fxmul.vv albh_v, a_v, bh_v, conf_mulus_r
	li	conf_mulsu_r, (FX_BU | FX_CLIP_PASS)
	fxmul.vv ahbl_v, ah_v, b_v, conf_mulsu_r

loop:
	lwai.v	a_v, aptr_r, fourvlr_r
	subu	len_r, MAXVLEN
 	sltu	done_r, len_r, MAXVLEN+1
	fxadd.vv res_v, albh_v, albl_v, conf_shladd_r
	##
	##
	srav.vs	ah_v, a_v, shift_r
	##	
	lwai.v	b_v, bptr_r, fourvlr_r
	##
	fxadd.vv res_v, ahbl_v, res_v, conf_shladd_r
	##
	##
	fxmul.vv albl_v, a_v, b_v, conf_muluu_r
	srav.vs	bh_v, b_v, shift_r
	swai.v	res_v, resptr_r, fourvlr_r
	##
	fxmul.vv albh_v, a_v, bh_v, conf_mulus_r
	##
	##
	##
	fxmul.vv ahbl_v, ah_v, b_v, conf_mulsu_r
	beqz	done_r, loop



/* Wind up after the loop */

	ctvu	len_r, vlr
	fxadd.vv res_v, albh_v, albl_v, conf_shladd_r
	fxadd.vv res_v, ahbl_v, res_v, conf_shladd_r
	sw.v	res_v, resptr_r

	j	ra

	.end	tor_mul_vi32vi32_vi32


#endif
