// $Header: /n/lasagna/db/davidj/spert/src/quicknet/H-spert0/../RCS/QN_SentDB.h,v 1.4 1996/03/01 03:37:40 davidj Exp $

#ifndef QN_SentDB_h_INCLUDED
#define QN_SentDB_h_INCLUDED

#include <QN_config.h>
#include <stddef.h>
#ifdef QN_HAVE_LIMITS_H
#include <limits.h>
#endif
#include "QN_types.h"

// QN_SentDB is a base class used to define an interface to a "Framed
// Sentence Database".  This database is one or more files of speech data that
// contain multiple sentences, with each sentence composed of multiple frames.
// For each frame there are multiple single precision floating-point features
// combined with zero or more integer labels and zero or more target values.
// An example of such a database is a classical pfile, as used with BoB - this
// may have 18 feature values, 1 label value and 0 target values.

class QN_SentDB
{
public:
    virtual ~QN_SentDB() {};

    // Some useful enums - particularly for constructors
    enum {
	ALL_FEATURES = INT_MAX,
	ALL_LABELS = INT_MAX,
	ALL_TARGETS = INT_MAX,
	ALL_SENTENCES = INT_MAX
    };

    // Return the number of features in a frame
    virtual size_t num_features_in_frame() = 0;
    // Return the number of labels in a frame
    virtual size_t num_labels_in_frame() = 0;
    // Return the number of targets in a frame
    virtual size_t num_targets_in_frame() = 0;

    // Return the total number of frames in the database
    virtual size_t num_frames() = 0;
    // Return the total number of sentences in the database
    virtual size_t num_sentences() = 0;

    // Return the number of frames in a range of sentences
    virtual size_t num_frames_in_sents(size_t first_sent,
				       size_t num_sents) = 0;

    // Gets the labels and features for a range of sentences.
    // If ftrs is NULL, do not get features.
    // If labs is NULL, do not get labels.
    // Use num_frames_in_sents() first to check space
    // Returns the number of frames
    virtual size_t get_sentences(size_t sentno, size_t num_sent,
				 QNUInt32* sentnos, QNUInt32* framenos,
				 float* ftrs, QNUInt32* labs,
				 float* tgts) = 0;
};

// #ifdef QN_SentDB_h_INCLUDED
#endif
