// $Header: /n/lasagna/db/davidj/spert/src/quicknet/RCS/QN_trn.h,v 1.9 1996/11/23 01:25:18 davidj Exp $

#ifndef QN_trn_h_INCLUDED
#define QN_trn_h_INCLUDED


#include <QN_config.h>
#include <assert.h>
#include <stddef.h>
#ifdef QN_HAVE_LIMITS_H
#include <limits.h>
#endif
#include "QN_Logger.h"
#include "QN_types.h"
#include "QN_streams.h"
#include "QN_RateSchedule.h"
#include "QN_MLP.h"

// A class for performing MLP training with hard targets.

class QN_HardSentTrainer
{
public:
    QN_HardSentTrainer(int a_debug, const char* a_dbgname,
		       int a_verbose, QN_MLP* a_mlp,
		       QN_InFtrStream* a_train_ftr_str,
		       QN_InLabStream* a_train_lab_str,
		       QN_InFtrStream* a_cv_ftr_str,
		       QN_InLabStream* a_cv_lab_str,
		       QN_RateSchedule* a_lr_sched,
		       float a_targ_low, float a_targ_high,
		       const char* a_wlog_template, size_t a_bunch_size);
    ~QN_HardSentTrainer();
    // Actually do training.
    void train();

protected:
    QN_ClassLogger clog;	// Logging class.
    int verbose;		// If non-zero, produce extra status msgs.
    QN_MLP* mlp;		// MLP being trained.
    const size_t mlp_inps;	// Number of inputs to the MLP.
    const size_t mlp_outs;	// Number of outputs from MLP.
    
    QN_InFtrStream* train_ftr_str; // Feature stream used for training.
    QN_InLabStream* train_lab_str; // Label stream used for training.
    QN_InFtrStream* cv_ftr_str;	// Feature stream used for cross validation.
    QN_InLabStream* cv_lab_str;	// Label stream used for cross validation.

    QN_RateSchedule* lr_sched;	// Learning rate schedule.
    const float targ_low;	// Value used for training 0 probability.
    const float targ_high;	// Value used for training 1 probability.

    const size_t bunch_size;	// Number of presentations to pass to net
				// at one time.
    const size_t inp_buf_size;	// Number of values passed to MLP.
    const size_t out_buf_size;	// Number of values output by MLP.
    const size_t targ_buf_size;	// Number of values in training target buffer.

    float* inp_buf;		// Buffer for values going into net.
    float* out_buf;		// Buffer for values coming out from net.
    float* targ_buf;		// Buffer for correct outputs when training.
    QNUInt32*lab_buf;		// One bunches worth of labels.

    char* wlog_template;	// Template used for temp weight file names.
    const int pid;		// Proces ID - used for building weight file
				// names.

    float learn_rate;		// The current learning rate.
    size_t epoch;		// Current epoch.

// Local functions.
    double cv_epoch();		// Do one epochs worth of cross validation.
    double train_epoch();	// Do one epochs worth of training.
    void set_learnrate();	// Set the learning rates in the net based
				// on the value of learn_rate.
    // Read weights from a file - should really be elsewhere!
    void read_weights(const char* wfile_name, QN_MLP* mlp);
    // Write weights to a file - should really be elsewhere!
    void write_weights(const char* wfile_name, QN_MLP* mlp);
};

// A class for performing MLP training with soft targets.

class QN_SoftSentTrainer
{
public:
    QN_SoftSentTrainer(int a_debug, const char* a_dbgname,
		       int a_verbose, QN_MLP* a_mlp,
		       QN_InFtrStream* a_train_ftr_str,
		       QN_InFtrStream* a_train_targ_str,
		       QN_InFtrStream* a_cv_ftr_str,
		       QN_InFtrStream* a_cv_targ_str,
		       QN_RateSchedule* a_lr_sched,
		       float a_targ_low, float a_targ_high,
		       const char* a_wlog_template, size_t a_bunch_size);
    ~QN_SoftSentTrainer();
    // Actually do training.
    void train();

protected:
    QN_ClassLogger clog;	// Logging class.
    int verbose;		// If non-zero, produce extra status msgs.
    QN_MLP* mlp;		// MLP being trained.
    const size_t mlp_inps;	// Number of inputs to the MLP.
    const size_t mlp_outs;	// Number of outputs from MLP.
    
    QN_InFtrStream* train_ftr_str; // Feature stream used for training.
    QN_InFtrStream* train_targ_str; // Label stream used for training.
    QN_InFtrStream* cv_ftr_str;	// Feature stream used for cross validation.
    QN_InFtrStream* cv_targ_str; // Label stream used for cross validation.

    QN_RateSchedule* lr_sched;	// Learning rate schedule.
    const float targ_low;	// Value used for training 0 probability.
    const float targ_high;	// Value used for training 1 probability.

    const size_t bunch_size;	// Number of presentations to pass to net
				// at one time.
    const size_t inp_buf_size;	// Number of values passed to MLP.
    const size_t out_buf_size;	// Number of values output by MLP.
    const size_t targ_buf_size;	// Number of values in training target buffer.

    float* inp_buf;		// Buffer for values going into net.
    float* out_buf;		// Buffer for values coming out from net.
    float* targ_buf;		// Buffer for correct outputs when training.

    char* wlog_template;	// Template used for temp weight file names.
    const int pid;		// Proces ID - used for building weight file
				// names.

    float learn_rate;		// The current learning rate.
    size_t epoch;		// Current epoch.

// Local functions.
    double cv_epoch();		// Do one epochs worth of cross validation.
    double train_epoch();	// Do one epochs worth of training.
    void set_learnrate();	// Set the learning rates in the net based
				// on the value of learn_rate.
    // Read weights from a file - should really be elsewhere!
    void read_weights(const char* wfile_name, QN_MLP* mlp);
    // Write weights to a file - should really be elsewhere!
    void write_weights(const char* wfile_name, QN_MLP* mlp);
};

// A class for performing MLP training with soft targets.

#endif
