#ifndef NO_RCSID
const char* qncopywts_rcsid = 
    "$Header: /n/abbott/da/drspeech/src/SPRACHcore/quicknet/RCS/qncopywts.cc,v 1.3 2001/04/16 13:41:26 dpwe Exp $";
#endif

/* Must include the config.h file first */
#include <QN_config.h>
#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef QN_HAVE_LIMITS_H
#include <limits.h>
#endif
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS (0)
#define EXIT_FAILURE (1)
#endif
#include <sys/types.h>
#ifdef QN_HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef QN_HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif


#ifndef FILENAME_MAX
#define FILENAME_MAX (MAXPATHLEN)
#endif
#ifdef QN_HAVE_GETHOSTNAME
extern "C" int gethostname(char* name, int len);
#endif

#include "fltvec.h"
#include "QuickNet.h"

#include "QN_MLPWeightFile_MLPW.h"

static struct {
    char* iwts_filename;
    char* iwts_format_str;
    int iwts_format_val;
    char* owts_filename;
    char* owts_format_str;
    int owts_format_val;
    int in2hid_exp;
    int hid2out_exp;
    int hidbias_exp;
    int outbias_exp;
    int dbg;			// Debug level.
    int verbose;
} config;

enum _QN_WTSFMT {
    QN_WTSFMT_NONE = 0,
    QN_WTSFMT_RAP3, 
    QN_WTSFMT_MLPW, 	// subtype unspecified
    QN_WTSFMT_MLPW_F, 
    QN_WTSFMT_MLPW_D,
    QN_WTSFMT_MLPW_L,
    QN_WTSFMT_MLPW_S,
    QN_WTSFMT_MLPW_C
};

void
set_defaults(void)
{
    config.iwts_filename = NULL;
    config.iwts_format_str = "RAP3";
    config.iwts_format_val = QN_WTSFMT_RAP3;
    config.owts_filename = NULL;
    config.owts_format_str = "RAP3";
    config.owts_format_val = QN_WTSFMT_RAP3;
    config.in2hid_exp = 2;
    config.hid2out_exp = 2;
    config.hidbias_exp = 4;
    config.outbias_exp = 4;
    config.dbg = 0;
    config.verbose = 1;
}

QN_ArgEntry argtab[] =
{
{ NULL, "Net weights file copying", QN_ARG_DESC },
{ "iwts_file", "Input weights file", QN_ARG_STR,
  &(config.iwts_filename), QN_ARG_REQ },
{ "iwts_format", "Input file format", QN_ARG_STR, 
  &(config.iwts_format_str) }, 
{ "owts_file", "Output weights file", QN_ARG_STR,
  &(config.owts_filename), QN_ARG_REQ },
{ "owts_format", "Output file format", QN_ARG_STR, 
  &(config.owts_format_str) }, 
{ "in2hid_exp", "Exponent value for fixed-point input to hidden weights", QN_ARG_INT, 
  &(config.in2hid_exp) }, 
{ "hid2out_exp", "Exponent value for fixed-point hidden to output weights", QN_ARG_INT, 
  &(config.hid2out_exp) }, 
{ "hidbias_exp", "Exponent value for fixed-point hidden units bias", QN_ARG_INT, 
  &(config.hidbias_exp) }, 
{ "outbias_exp", "Exponent value for fixed-point output units bias", QN_ARG_INT, 
  &(config.outbias_exp) }, 
{ "debug", "Output additional diagnostic information",
  QN_ARG_INT, &(config.dbg) },
{ "verbose", "Output diagnostic information",
  QN_ARG_BOOL, &(config.verbose) },
{ NULL, NULL, QN_ARG_NOMOREARGS }
};

static int wts_fmt_str2val(char *str, int *val) {
    // convert wts fmt strings into vals, or error (return 0)
    if (strcmp(str, "RAP3") == 0) *val = QN_WTSFMT_RAP3;
    else if (strcmp(str, "MLPW") == 0) *val = QN_WTSFMT_MLPW;
    else if (strcmp(str, "MLPWF") == 0) *val = QN_WTSFMT_MLPW_F;
    else if (strcmp(str, "MLPWD") == 0) *val = QN_WTSFMT_MLPW_D;
    else if (strcmp(str, "MLPWL") == 0) *val = QN_WTSFMT_MLPW_L;
    else if (strcmp(str, "MLPWS") == 0) *val = QN_WTSFMT_MLPW_S;
    else if (strcmp(str, "MLPWC") == 0) *val = QN_WTSFMT_MLPW_C;
    else {
	fprintf(stderr, "qncopywts: '%s' is not a known wts fmt spec (RAP3/MLPW/MLPWx, x=F/D/L/S/C)\n", str);
	return 0;
    }
    return 1;
}

static void
copy_run(FILE* out_file, FILE* logfile, int verbose)
{
    if (verbose>0) {
	// A note for the logfile.
        time_t now = time(NULL);
	fprintf(logfile, "\nProgram version: qncopywts %s.\n", QN_VERSION);
#ifdef QN_HAVE_GETHOSTNAME
	char hostname[MAXHOSTNAMELEN];
	assert(gethostname(hostname, MAXHOSTNAMELEN)==0);
	fprintf(logfile, "Host: %s\n", hostname);
#endif
	fprintf(logfile, "Program start: %s", ctime(&now));
    }
    if (verbose>0) {
	fprintf(logfile, "Opening wts file...\n");
    }

    // Open input file
    FILE* init_weight_fp = NULL;
    const char* init_weight_file = config.iwts_filename;
    if (strcmp(init_weight_file, "")==0)
    {
	QN_ERROR(NULL, "Null input weightfile specified.");
    }
    else
	init_weight_fp = QN_open(init_weight_file, "r");
    float min, max;
    if (verbose>0)
    {
	fprintf(logfile,"Loading weights...\n");
    }
    QN_MLPWeightFile *inwfile;
    QN_MLPWeightFile *outwfile;

    int debug = config.dbg;

    if (config.iwts_format_val == QN_WTSFMT_RAP3) {
	inwfile = new QN_MLPWeightFile_RAP3(debug, init_weight_fp,
				  QN_READ,
				  init_weight_file, 0,0,0);
    } else if (config.iwts_format_val == QN_WTSFMT_MLPW) {
	inwfile = new QN_MLPWeightFile_MLPW(debug, init_weight_fp,
				  QN_READ,
				  init_weight_file, 0, NULL);
    } else {
	fprintf(stderr, "Input format must be RAP3 or MLPW (no subtypes)\n");
	exit(1);
    }

    if (inwfile->num_layers() != 3) {
	fprintf(stderr, "qncopywts: I handle only 3 layer MLPs, not %d layers\n", inwfile->num_layers());
	exit(1);
    }

    int mlp3_input_size = inwfile->size_layer(QN_LAYER0);
    int mlp3_hidden_size = inwfile->size_layer(QN_LAYER1);
    int mlp3_output_size = inwfile->size_layer(QN_LAYER2);
    QN_OutputLayerType mlp3_output_type = QN_OUTPUT_SOFTMAX;
    QN_MLPW_ntype ntype = QN_MLPW_nt_softmax;

    // Create the MLP.
    QN_MLP* mlp = new QN_MLP_OnlineFl3(debug, "mlp",
		  mlp3_input_size, mlp3_hidden_size, mlp3_output_size,
		  mlp3_output_type);

    QN_read_weights(*inwfile, *mlp, &min, &max, debug);
    QN_OUTPUT("Weights loaded from '%s', min=%g max=%g.",
	      init_weight_file, min, max);

    // So write them
    int lsz[3];
    lsz[0] = mlp3_input_size;
    lsz[1] = mlp3_hidden_size;
    lsz[2] = mlp3_output_size;
    QN_MLPWeightFile_MLPW *outmlpwfile;

    if (config.owts_format_val == QN_WTSFMT_RAP3) {
	outwfile = new QN_MLPWeightFile_RAP3(debug, out_file, 
				  QN_WRITE,
				  init_weight_file, mlp3_input_size, mlp3_hidden_size, mlp3_output_size);
    } else if (config.owts_format_val == QN_WTSFMT_MLPW_F) {
	outmlpwfile = new QN_MLPWeightFile_MLPW(debug, out_file,
					     QN_WRITE,
					     init_weight_file, 3, lsz, 
					     ntype);
	outmlpwfile->set_datatype(QN_MLP3_INPUT2HIDDEN, QN_MLPW_dt_float);
	outmlpwfile->set_datatype(QN_MLP3_HIDDEN2OUTPUT, QN_MLPW_dt_float);
	outmlpwfile->set_datatype(QN_MLP3_HIDDENBIAS, QN_MLPW_dt_float);
	outmlpwfile->set_datatype(QN_MLP3_OUTPUTBIAS, QN_MLPW_dt_float);
	outwfile = outmlpwfile;
    } else if (config.owts_format_val == QN_WTSFMT_MLPW_D) {
	outmlpwfile = new QN_MLPWeightFile_MLPW(debug, out_file,
					     QN_WRITE,
					     init_weight_file, 3, lsz, 
					     ntype);
	outmlpwfile->set_datatype(QN_MLP3_INPUT2HIDDEN, QN_MLPW_dt_double);
	outmlpwfile->set_datatype(QN_MLP3_HIDDEN2OUTPUT, QN_MLPW_dt_double);
	outmlpwfile->set_datatype(QN_MLP3_HIDDENBIAS, QN_MLPW_dt_double);
	outmlpwfile->set_datatype(QN_MLP3_OUTPUTBIAS, QN_MLPW_dt_double);
	outwfile = outmlpwfile;
    } else if (config.owts_format_val == QN_WTSFMT_MLPW_L) {
	outmlpwfile = new QN_MLPWeightFile_MLPW(debug, out_file,
					     QN_WRITE,
					     init_weight_file, 3, lsz, 
					     ntype);
	outmlpwfile->set_datatype(QN_MLP3_INPUT2HIDDEN, QN_MLPW_dt_int32,
				  config.in2hid_exp);
	outmlpwfile->set_datatype(QN_MLP3_HIDDEN2OUTPUT, QN_MLPW_dt_int32,
				  config.hid2out_exp);
	outmlpwfile->set_datatype(QN_MLP3_HIDDENBIAS, QN_MLPW_dt_int32,
				  config.hidbias_exp);
	outmlpwfile->set_datatype(QN_MLP3_OUTPUTBIAS, QN_MLPW_dt_int32, 
				  config.outbias_exp);
	outwfile = outmlpwfile;
    } else if (config.owts_format_val == QN_WTSFMT_MLPW_S) {
	outmlpwfile = new QN_MLPWeightFile_MLPW(debug, out_file,
					     QN_WRITE, 
					     init_weight_file, 3, lsz, 
					     ntype);
	outmlpwfile->set_datatype(QN_MLP3_INPUT2HIDDEN, QN_MLPW_dt_int16,
				  config.in2hid_exp);
	outmlpwfile->set_datatype(QN_MLP3_HIDDEN2OUTPUT, QN_MLPW_dt_int16,
				  config.hid2out_exp);
	// biases held at higher resolution
	outmlpwfile->set_datatype(QN_MLP3_HIDDENBIAS, QN_MLPW_dt_int32,
				  config.hidbias_exp);
	outmlpwfile->set_datatype(QN_MLP3_OUTPUTBIAS, QN_MLPW_dt_int32, 
				  config.outbias_exp);
	outwfile = outmlpwfile;
    } else if (config.owts_format_val == QN_WTSFMT_MLPW_C) {
	outmlpwfile = new QN_MLPWeightFile_MLPW(debug, out_file,
					     QN_WRITE,
					     init_weight_file, 3, lsz, 
					     ntype);
	outmlpwfile->set_datatype(QN_MLP3_INPUT2HIDDEN, QN_MLPW_dt_int8,
				  config.in2hid_exp);
	outmlpwfile->set_datatype(QN_MLP3_HIDDEN2OUTPUT, QN_MLPW_dt_int8,
				  config.hid2out_exp);
	// biases held at higher resolution
	outmlpwfile->set_datatype(QN_MLP3_HIDDENBIAS, QN_MLPW_dt_int16,
				  config.hidbias_exp);
	outmlpwfile->set_datatype(QN_MLP3_OUTPUTBIAS, QN_MLPW_dt_int16, 
				  config.outbias_exp);
	outwfile = outmlpwfile;
    }

    QN_write_weights(*outwfile, *mlp, &min, &max, debug);
    QN_OUTPUT("Weights written to '%s', min=%g max=%g.",
	      config.owts_filename, min, max);

    // Close the output file
    delete outwfile;

    // clean up input
    delete inwfile;
    fclose(init_weight_fp);

    // get rid of mlp
    delete mlp;

    if (verbose) {
	// A note for the logfile.
        time_t now = time(NULL);
	fprintf(logfile, "Program stop: %s", ctime(&now));
    }
}

int
main(int argc, const char* argv[])
{
    char* progname;		// The name of the program - set by QN_args.
    FILE* out;
    FILE* logfile;
    char logbuf[160];

    set_defaults();
    QN_initargs(&argtab[0], &argc, &argv, &progname);

    // map wts_formats to vals
    if(wts_fmt_str2val(config.iwts_format_str, &config.iwts_format_val)==0) {
	fprintf(stderr, "error setting iwts fmt\n");
	exit(-1);
    }
    // Can't specify MLPW subtype on input
    if (config.owts_format_val == QN_WTSFMT_MLPW_F \
	|| config.owts_format_val == QN_WTSFMT_MLPW_D \
	|| config.owts_format_val == QN_WTSFMT_MLPW_L \
	|| config.owts_format_val == QN_WTSFMT_MLPW_S \
	|| config.owts_format_val == QN_WTSFMT_MLPW_C) {
	fprintf(stderr, "No MLPW subtypes on input - just say MLPW\n");
	exit(-1);
    }
    if(wts_fmt_str2val(config.owts_format_str, &config.owts_format_val)==0) {
	fprintf(stderr, "error setting owts fmt\n");
	exit(-1);
    }
    // Default MLPW output is MLPW_F
    if (config.owts_format_val == QN_WTSFMT_MLPW) {
	config.owts_format_val = QN_WTSFMT_MLPW_F;
    }

    logfile = stderr;

    // Enable line buffering on log file.
    // *** Note: there is a bug with newlib-1.6.1 (as used on SPERT) that
    // *** makes setvbuf fail if the second argument is NULL.
    if (setvbuf(logfile, logbuf, _IOLBF, sizeof(logbuf)))
	abort();		// Fail on error
    if (config.verbose) {
	QN_printargs(logfile, progname, &argtab[0]);
	QN_logger = new QN_Logger_Simple(logfile, stderr, progname);
    } else {
	logfile = NULL;
	QN_logger = new QN_Logger_Simple(logfile, stderr, progname);
    }
    // Ensure the lines are sent to the logfile as soon as they are printed.

    out = fopen(config.owts_filename, "w");
    if (out==NULL) {
	QN_ERROR(NULL, "failed to open wts file %s for writing.",
		 config.owts_filename);
    }

    // Do the transfer
    copy_run(out, logfile, config.verbose);

    if (fclose(out)) {
	QN_ERROR(NULL, "failed to close output file %s.",
		 config.owts_filename);
    }

    exit(EXIT_SUCCESS);
}
