const char* qnforward_rcsid =
    "$Header: /u/davidj/spert/src/quicknet/RCS/qnforward.cc,v 1.76 1997/05/21 22:07:44 davidj Exp $";
//
// A simple forward pass (i.e. recognition only) MLP program.

/* Must include the config.h file first */
#include <QN_config.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#ifdef QN_HAVE_LIMITS_H
#include <limits.h>
#endif
#include <time.h>
#ifdef QN_HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef QN_HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef __GNUC__
extern "C" {
    typedef void (*new_handler)(void);
    new_handler set_new_handler (new_handler);
}
#endif
#ifndef FILENAME_MAX
#define FILENAME_MAX 128
#endif
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1
#endif

#ifdef QN_HAVE_GETHOSTNAME
extern "C" int gethostname(char* name, int len);
#endif

#include "fltvec.h"
#include "QuickNet.h"


struct {
    char* ftrfile_filename;		// Input PFile name.
    char* ftrfile_format;	// Format of input feature file.
    int ftrfile_num_ftrs;	// Number of features
    char* weights_filename;	// Input weight file name.
    char* norms_filename;	// Input norm file name.
    char* output_filename;	// Output file name.
    char* output_format;	// Output file format.
    int n_input;		// Number of units in input layer.
    int n_hidden;		// Number of units in hidden layer.
    int n_output;		// Number of units in output layer.
    char* mlp_output_type;
    int weight_exp;		// Fixed point net weight exponent.
    int window_width;
    int window_offset;
    int first_ftr;		// First feature to use.
    int num_ftrs;		// Number of features to use.
    int first_sent;		// First sentence to recognize.
    int num_sents;		// Number of sentences to recognize.
    int use_fixed;		// 1 if to use a fixed point net.
    int verbose;		// 1 if extra status messages.
    int dbg;			// Debug level.
} config;

void
set_defaults(void)
{
    config.ftrfile_filename = "(unset)";
    config.ftrfile_format = "pfile";
    config.ftrfile_num_ftrs = 0;
    config.weights_filename = "(unset)";
    config.norms_filename = "";
    config.output_filename = "-"; // Output file name
    config.output_format = "rapascii"; // Output file format
    config.n_input = 153;	// Number of units in input layer
    config.n_hidden = 200;	// Number of units in hidden layer
    config.n_output = 56;	// Number of units in output layer
    config.mlp_output_type = "softmax";
    config.weight_exp = 2;
    config.window_width = 9;
    config.window_offset = -4;
    config.first_ftr = 1;	// First feature to use
    config.num_ftrs = 0;	// Number of features to use
    config.first_sent = 0;	// First sentence to recognize
    config.num_sents = INT_MAX;	// Number of sentences to recognize
    config.use_fixed = 0;
    config.verbose = 1;
    config.dbg = 0;
}

QN_ArgEntry argtab[] =
{
{ NULL, "Fast phoneme recognition MLP forward pass\n", QN_ARG_DESC},
{ "ftrfile", "Feature file", QN_ARG_STR,
      &(config.ftrfile_filename), QN_ARG_REQ},
{ "ftrfile_format", "Feature file format (pfile,pre)", QN_ARG_STR,
      &(config.ftrfile_format) },
{ "ftrfile_num_ftrs", "Width of feature file", QN_ARG_INT,
      &(config.ftrfile_num_ftrs) },
{ "input_weightfile", "Input weight file", QN_ARG_STR,
  &(config.weights_filename), QN_ARG_REQ},
{ "input_normfile", "Norm file", QN_ARG_STR,
  &(config.norms_filename) },
{ "output_file", "Output file [default stdout]", QN_ARG_STR,
  &(config.output_filename) },
{ "output_file_format", "Output format (rapascii,raphex,rapbin,lna8) [default rapascii]",
  QN_ARG_STR, &(config.output_format) },
{ "window_width", "Number if frames in input window",
  QN_ARG_INT, &(config.window_width) },
{ "window_offset", "Offset of input window", QN_ARG_INT,
  &(config.window_offset) },
{ "num_input", "Number of units in input layer",
  QN_ARG_INT, &(config.n_input) },
{ "num_hidden","Number of units in hidden layer",
  QN_ARG_INT, &(config.n_hidden) },
{ "num_output","Number of units in output layer",
  QN_ARG_INT, &(config.n_output) },
{ "mlp_output_type","Type of non-linearity in MLP output layer",
  QN_ARG_STR, &(config.mlp_output_type) },
{ "mlp_weightexp","Fixed point net weight exponent",
  QN_ARG_INT, &(config.weight_exp) },
{ "first_sent", "Number of first sentence",
  QN_ARG_INT, &(config.first_sent) },
{ "num_sents", "Number of sentences",
  QN_ARG_INT, &(config.num_sents) },
{ "first_ftr", "First feature used from database",
  QN_ARG_INT, &(config.first_ftr) },
{ "num_ftrs", "Number of features used from database",
  QN_ARG_INT, &(config.num_ftrs) },
{ "use_fx", "Use a fixed point net",
  QN_ARG_BOOL, &(config.use_fixed) },
{ "verbose", "Output extra status messages",
  QN_ARG_BOOL, &(config.verbose) },
{ "debug", "Output additional diagnostic information",
  QN_ARG_INT, &(config.dbg) },
{ NULL, NULL, QN_ARG_NOMOREARGS }
};



void
forward(FILE* db_file, QN_StreamType ftrfile_type,
	FILE* weights_file, FILE* norms_file,
        FILE* output_file, int verbose)
{
    enum {
	BATCH_SIZE = 16
    };
    time_t now;

    time(&now);
    QN_OUTPUT("Program version: qnforward version %s.",
	      QN_VERSION);
#ifdef QN_HAVE_GETHOSTNAME
    {
	int ec;
	char hostname[MAXHOSTNAMELEN];
	ec = gethostname(hostname, MAXHOSTNAMELEN);
	if (ec!=0)
	    QN_ERROR(NULL, "gethostname failed - %s.", strerror(errno));
	QN_OUTPUT("Host: %s", hostname);
    }
#endif
    QN_OUTPUT("Program start: %.24s", ctime(&now));
    if (verbose)
	QN_OUTPUT("Opening feature file...");

    // Get the combined feature and label stream.
    QN_InFtrLabStream* ftrlab_str; // The feature and label input stream.
    switch(ftrfile_type)
    {
    case QN_STREAM_PFILE:
    {
	QN_InFtrLabStream_PFile* pfile_str =
	    new QN_InFtrLabStream_PFile(config.dbg, // Select debugging.
					"input", // Debugging tag.
					db_file, // Feature file.
					1 // Indexed flag.
		);
	if (config.ftrfile_num_ftrs!=0
	    && pfile_str->num_ftrs()!=(size_t) config.ftrfile_num_ftrs)
	{
	    QN_ERROR(NULL, "Feature file has %lu features, command line"
		     " specified %lu features.",
		     (unsigned long) pfile_str->num_ftrs(),
		     config.ftrfile_num_ftrs);
	}
	ftrlab_str = pfile_str;
	break;
    }
    case QN_STREAM_PREFILE:
    {
	QN_InFtrLabStream_PreFile* prefile_str =
	    new QN_InFtrLabStream_PreFile(config.dbg, // Select debugging.
					  "input", // Debugging tag.
					  db_file,	// Feature file.
					  config.ftrfile_num_ftrs, // No. ftrs.
					  1 // Indexed
		);
	ftrlab_str = prefile_str;
	break;
    }
    default:
	ftrlab_str = NULL;
	assert(0);
    }
    size_t ftrlab_num_sents = ftrlab_str->num_segs();
    size_t ftrlab_num_frames = ftrlab_str->num_frames();
    size_t ftrlab_num_ftrs = ftrlab_str->num_ftrs();
    QN_OUTPUT("Feature file opened - %lu sentences, %lu frames, "
	      "%lu features in each frame.",
	      (unsigned long) ftrlab_num_sents,
	      (unsigned long) ftrlab_num_frames,
	      (unsigned long) ftrlab_num_ftrs);

    // Split the combined feature and label stream into separate,
    // asynchronous streams.
    QN_InFtrStream_SplitFtrLab* split_ftr_str =
	new QN_InFtrStream_SplitFtrLab(config.dbg, "split_ftr",
				       *ftrlab_str, 500);
    QN_InLabStream_SplitFtrLab* split_lab_str =
	new QN_InLabStream_SplitFtrLab(*split_ftr_str);
    QN_InFtrStream* ftr_str = split_ftr_str;
    QN_InLabStream* lab_str = split_lab_str;
    size_t num_ftrs = ftr_str->num_ftrs();
    size_t num_labs = lab_str->num_labs();
    assert(num_labs==1);

    // If a normalization file is used, normalize the feature stream.
    QN_InFtrStream_Norm* norm_str = NULL;
    if (norms_file!=NULL)
    {
	float* norm_mean = new float[num_ftrs];
	float* norm_sdevs = new float[num_ftrs];
	float* norm_bias = new float[num_ftrs];
	float* norm_scale = new float[num_ftrs];
	if (QN_read_norms_rap(norms_file, num_ftrs,
			      norm_mean, norm_sdevs))
	{
	    QN_ERROR(NULL, "failed to read norm file.");
	}
	// We actually want to subtract the mean and divide by the standard
	// deviation - the "Norm" filter adds a bias and multiplies by a
	// scale.
	neg_vf_vf(num_ftrs, norm_mean, norm_bias);
	recip_vf_vf(num_ftrs, norm_sdevs, norm_scale);
	norm_str = new QN_InFtrStream_Norm(config.dbg, "norm", *ftr_str,
					   norm_bias, norm_scale);
	ftr_str = norm_str;
    }

    // If we do not want all the features, extract just the ones we need.
    QN_InFtrStream_Narrow* narrow_str = NULL;
    if (config.first_ftr!=0
	|| ((size_t) config.num_ftrs!=num_ftrs && config.num_ftrs!=0)
	)
    {
	size_t narrow_num_ftrs;
	if (config.num_ftrs==0)
	    narrow_num_ftrs = QN_ALL;
	else
	    narrow_num_ftrs = config.num_ftrs;
	narrow_str = new QN_InFtrStream_Narrow(config.dbg, "narrow",
					       *ftr_str, config.first_ftr,
					       narrow_num_ftrs, 500);
	ftr_str = narrow_str;
    }

    // Extract just the sentences we want.
    QN_InFtrStream_Cut* cut_ftr_str = NULL;
    QN_InLabStream_Cut* cut_lab_str = NULL;
    if (config.first_sent!=0 || config.num_sents!=INT_MAX)
    {
	size_t num_sents;
	if (config.num_sents==INT_MAX)
	    num_sents = QN_ALL;
	else
	    num_sents = config.num_sents;
	cut_ftr_str = new QN_InFtrStream_Cut(config.dbg, "cutftr", *ftr_str,
					     config.first_sent, num_sents);
	cut_lab_str = new QN_InLabStream_Cut(config.dbg, "cutlab", *lab_str,
					     config.first_sent, num_sents);
	ftr_str = cut_ftr_str;
	lab_str = cut_lab_str;
    }


    // Window the input and label streams.
    QN_InFtrStream_SeqWindow* netinp_str = 
	new QN_InFtrStream_SeqWindow(config.dbg, // Level of debugging.
				     "ftr_win",	// Debugging tag.
				     *ftr_str, // Input stream.
				     config.window_width, // Length of window.
				     0,	// Top margin.
				     0, // Bottom margin.
				     100 // Buffer size.
				     );
				
    size_t lab_top_margin = -config.window_offset;
    size_t lab_bot_margin = config.window_width - lab_top_margin - 1;
    QN_InLabStream* corlabs_str = 
	new QN_InLabStream_SeqWindow(config.dbg,// Level of debugging.
				     "lab_win",	// Debugging tag.
				     *lab_str,  // Input label stream.
				     1,	        // Window length.
				     lab_top_margin,
				     lab_bot_margin,
				     100        // buffer size.
				     );

    // Here we generate and MLP and initialize it
    QN_OutputLayerType outlayer_type;
    if (strcmp(config.mlp_output_type, "sigmoid")==0)
	outlayer_type = QN_OUTPUT_SIGMOID;
    else if (strcmp(config.mlp_output_type, "sigmoidx")==0)
	outlayer_type = QN_OUTPUT_SIGMOID_XENTROPY;
    else if (strcmp(config.mlp_output_type, "softmax")==0)
	outlayer_type = QN_OUTPUT_SOFTMAX;
    else
    {
	QN_ERROR(NULL, "unknown output unit type '%s'.",
		 config.mlp_output_type);
	outlayer_type = QN_OUTPUT_SIGMOID;
    }
    QN_MLP* mlp_p;
    if (config.use_fixed)
    {
#ifdef QN_HAVE_LIBFXLIB
	mlp_p = new QN_MLP_OnlineFx3(config.dbg, // Debug level.
				     config.n_input,    // Input layers.
				     config.n_hidden,   // Hidden layers.
				     config.n_output,   // Output layers.
				     outlayer_type,     // Output layer type.
				     config.weight_exp, // in2hid exponent.
				     config.weight_exp  // hid2out exponent.
				     );
// Checking for sysspert() is a hacky way to check to see if
// we are actualy running on a SPERT board.
#ifndef QN_HAVE_SYSSPERT
	QN_WARN(NULL, "fixed point is slow on non-SPERT hosts.");
#endif
#else
	QN_ERROR(NULL, "fixed point not supported in this version.");
	mlp_p = NULL;
#endif
    }
    else
    {
	// Here we can use the bunch mode class for the forward
	// pass - the difference between bunch and online is only
	// in training, and the bunch mode class is faster.
	mlp_p = new QN_MLP_BunchFl3(config.dbg, // Debug level,
				    "fwdmlp", // Debugging name.
				    config.n_input, // Input layers
                                    config.n_hidden, // Hidden layers
				    config.n_output, // Output layers
				    outlayer_type,   // Output layer type
				    BATCH_SIZE
				    );
#ifdef QN_HAVE_SYSSPERT
	QN_WARN(NULL, "floating point is very slow on SPERT.");
#endif
    }
    QN_MLP& mlp = *mlp_p;
    if (config.verbose)
    {
	QN_OUTPUT("Loading weights...");
    }
    float min, max;
    QN_MLPWeightFile_RAP3 wfile(config.dbg, weights_file, QN_READ,
				config.weights_filename,
				config.n_input, config.n_hidden,
				config.n_output);
    QN_read_weights(wfile, mlp, &min, &max, 0);
    QN_OUTPUT("Loaded weights and biases, "
	      "min=%f max=%f.", min, max);
    // Now we actually do the work
    if (config.verbose)
    {
	QN_OUTPUT("Starting forward pass.");
    }

    // Send the output to a RAP-style probability file
    enum QN_StreamType format; // The format of the output file
    if (strcmp(config.output_format, "rapascii")==0)
	format = QN_STREAM_RAPACT_ASCII;
    else if (strcmp(config.output_format, "raphex")==0)
	format = QN_STREAM_RAPACT_HEX;
    else if (strcmp(config.output_format, "rapbin")==0)
	format = QN_STREAM_RAPACT_BIN;
    else if (strcmp(config.output_format, "lna8")==0)
	format = QN_STREAM_LNA8;
    else
    {
	QN_ERROR(NULL, "unknown output format '%s'.", config.output_format);
	format = QN_STREAM_RAPACT_ASCII; // Avoid warning
    }
    QN_OutFtrStream* outstream_p;
    switch(format)
    {
    case QN_STREAM_RAPACT_ASCII:
    case QN_STREAM_RAPACT_HEX:
    case QN_STREAM_RAPACT_BIN:
	outstream_p =
	    new QN_OutFtrStream_Rapact(config.dbg,	// Debug.
				       "outstream",	// Instance name.
				       output_file,     // output stream.
				       config.n_output, // Number of units.
				       format,   // Which particular format.
				       1 );	     // Whether online or not.
	break;
    case QN_STREAM_LNA8:
	outstream_p = new QN_OutFtrStream_LNA8(config.dbg, "outstream",
					       output_file,
					       config.n_output );
#ifdef QN_HAVE_SYSSPERT
	QN_WARN(NULL, "LNA output is very slow on SPERT.");
#endif
	break;
    default:
	outstream_p = NULL;
	assert(0);
    }
    // Do the forward pass.
    QN_hardForward(config.dbg,	// Level of debugging.
		   "fwd",	// A name for this object.
		   config.verbose,	// Level of status reporting.
		   &mlp,	// The MLP to use.
		   netinp_str,	// The stream containing the windowd net input.
		   corlabs_str,	// The stream containing the corrct label.
		   outstream_p,	// The net output stream.
		   NULL,	// The label output stream.
		   BATCH_SIZE	// The number of frames in a bunch.
	);

    delete corlabs_str;
    delete netinp_str;
    if (cut_lab_str!=NULL)
	delete cut_lab_str;
    if (cut_ftr_str!=NULL)
	delete cut_ftr_str;
    if (narrow_str!=NULL)
	delete narrow_str;
    if (norm_str!=NULL)
	delete norm_str;
    delete split_lab_str;
    delete split_ftr_str;
    delete ftrlab_str;

    // All done
    time(&now);
    if (config.verbose)
    {
	QN_OUTPUT("Program stop: %.24s", ctime(&now));
    }
}

int
main(int argc, const char** argv)
{
    char* progname;		// The name of the program - set by icsiargs
    FILE* db_file;		// PFile stream
    FILE* weights_file;		// Weight file stream
    FILE* norms_file;		// Norm file stream
    FILE* output_file;		// Output file
    FILE* logfile;
    QN_StreamType db_streamtype; // The file type of the database
    char logbuf[160];		// A buffer for logging messages
    int stdout_is_outfile;	// 1 if stdout is used for activation output

    set_defaults();
    QN_initargs(&argtab[0], &argc, &argv, &progname);

    // Sort out logging stream
    if (strcmp(config.output_filename, "")==0
	|| strcmp(config.output_filename,"-")==0)
    {
	stdout_is_outfile = 1;
	logfile = stderr;
    }
    else
    {
	stdout_is_outfile = 0;
	logfile = stdout;
    }
    // Enable line buffering on log file
    // *** Note: there is a bug with newlib-1.6.1 (as used on SPERT) that
    // *** makes setvbuf fail if the second argument is NULL
    if (setvbuf(logfile, logbuf, _IOLBF, sizeof(logbuf)))
	abort();		// Fail on error
    QN_printargs(logfile, progname, &argtab[0]);
    QN_logger = new QN_Logger_Simple(logfile, stderr, progname);

    // Install our own out-of-memory handler if possible.
    // This is important on SPERT, as for some versions of spertdev the
    // default out of memory handler produces no output and is consequently
    // confusing.
#ifdef __GNUC__
    set_new_handler(QN_new_handler);
#endif

    if (config.window_width<0 || config.window_width>1000)
    {
	QN_ERROR(NULL, "window must be in range 0-1000.");
    }
    // Work out how many features in the stream.
    size_t net_num_ftrs;
    if (config.num_ftrs!=0)
	net_num_ftrs = config.num_ftrs;
    else if (config.ftrfile_num_ftrs!=0)
	net_num_ftrs = config.ftrfile_num_ftrs - config.first_ftr;
    else
	net_num_ftrs = QN_ALL;
    if (net_num_ftrs!=QN_ALL &&
	(size_t) config.window_width*net_num_ftrs != (size_t) config.n_input)
    {
	QN_ERROR(NULL, "number of input features (%i) "
		 "* window size (%i) "
		 "must equal input layer size (%i).",
		 (int) net_num_ftrs, (int) config.window_width,
		 (int) config.n_input);
    }

    if (strcmp(config.ftrfile_format, "pfile")==0)
    {
	db_streamtype = QN_STREAM_PFILE;
    }
    else if (strcmp(config.ftrfile_format, "pre")==0)
    {
	db_streamtype = QN_STREAM_PREFILE;
    }
    else
    {
	QN_ERROR(NULL, "Unknown file format '%s'.", config.ftrfile_format);
	db_streamtype = QN_STREAM_UNKNOWN;
    }
    db_file = fopen(config.ftrfile_filename, "r");
    if (db_file==NULL)
    {
	QN_ERROR(NULL, "failed to open %s.", config.ftrfile_filename);
    }
    weights_file = fopen(config.weights_filename, "r");
    if (weights_file==NULL)
    {
	QN_ERROR(NULL, "failed to open %s.", config.weights_filename);
    }
    if (strcmp(config.norms_filename, "")!=0)
    {
	norms_file = fopen(config.norms_filename, "r");
	if (norms_file==NULL)
	{
	    QN_ERROR(NULL, "failed to open %s.", config.norms_filename);
	}
    }
    else
	norms_file = NULL;
    if (stdout_is_outfile)
    {
	output_file = stdout;
    }
    else
    {
	output_file = fopen(config.output_filename, "w");
	if (output_file==NULL)
	{
	    QN_ERROR(NULL, "failed to open %s for output.",
		     config.output_filename);
	}
    }
    // Use bigger buffers for IO - especially important when using NFS
    // *** Note: there is a bug with newlib-1.6.1 (as used on SPERT) that
    // *** makes setvbuf fail if the second argument is NULL
    enum { DB_BUF_SIZE = 0x4000, OUTPUT_BUF_SIZE = 0x4000 };
    char* db_buffer = new char[DB_BUF_SIZE];
    setvbuf(db_file, db_buffer, _IOFBF, DB_BUF_SIZE);
    char* output_buffer = new char[OUTPUT_BUF_SIZE];
    setvbuf(output_file, output_buffer, _IOFBF, OUTPUT_BUF_SIZE);

    forward(db_file, db_streamtype, weights_file, norms_file,
	    output_file, config.verbose);

    if (fclose(db_file))
    {
	QN_ERROR(NULL, "failed to close feature file %s.",
		 config.ftrfile_filename);
    }
    delete db_buffer;
    if (fclose(weights_file))
    {
	QN_ERROR(NULL, "failed to close weights %s.", config.weights_filename);
    }
    if (fclose(norms_file))
    {
	QN_ERROR(NULL, "failed to close norms %s.", config.norms_filename);
    }
    if (output_file!=stdout)
    {
	if (fclose(output_file))
	{
	    QN_ERROR(NULL, "failed to close norms %s.", config.norms_filename);
	}
    }
    delete output_buffer;
    QN_FINISH();
}
