#ifndef NO_RCSID
const char* qnspc_multispertfx3_1_rcsid = 
    "$Header: /n/zinfandel/xa/fosler/quicknet/src/quicknet-v0_96-tetra/qnspc_multispertfx3_1.cc,v 1.2 1998/08/22 06:20:38 davidj Exp $";
#endif

#include <QN_config.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <float.h>
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS (0)
#define EXIT_FAILURE (1)
#endif
#include <sys/types.h>

#if defined(QN_HAVE_LIBSPCSLAVE) && defined(QN_HAVE_LIBFXLIB)

#include "fltvec.h"
extern "C" {
#include "fxlib.h"
}
#include "spcslave.h"
#include "QN_types.h"
#include "QN_Logger.h"
#include "QN_SPC_MultiSPERTFx3_1.h"

extern "C"
fxbool        mulnt_srx_mi16mi16_i_mi32t(fxsize_t m,
					 fxsize_t k,
					 fxsize_t n,
					 fxsize_t a_stride,
					 fxsize_t b_stride,
					 fxsize_t c_stride,
					 const fxint16* a,
					 const fxint16* b,
					 fxexp_t rshift,
					 fxint32* c);

extern "C"
fxbool multn_srx_srx_satadd_mi16mi16_i_i_mhi32_mhi32st(
					 fxsize_t m,
					 fxsize_t k,
					 fxsize_t n,
					 fxsize_t a_stride,
					 fxsize_t b_stride,
					 fxsize_t c_in_stride,
					 fxsize_t c_out_stride,
					 const fxint16* a,
					 const fxint16* b,
					 fxexp_t mul_rshift,
					 fxexp_t add_rshift,
					 const fxint16* c_in_hi,
					 const fxint16* c_in_lo,
					 fxint16* c_out_hi,
					 fxint16* c_out_lo);

extern "C" fxbool mul_srx_mi16mi16_i_mi32t(  fxsize_t m,
					 fxsize_t k,
					 fxsize_t n,
					 fxsize_t a_stride,
					 fxsize_t b_stride,
					 fxsize_t c_stride,
					 const fxint16* a,
					 const fxint16* b,
					 fxexp_t rshift,
					 fxint32* c);

// *** NOTE ***
// This program is NOT a normal QuickNet program.  It runs on the SPERT
// under the control of a host application, using SPC operations
// to receive requests for function invocation.

class QN_MLP_MultiSPERTFx3_slave
{
public:
    QN_MLP_MultiSPERTFx3_slave(int debug,       // Level of debugging
		     size_t n_input,  // Size of input layer
		     size_t n_hidden, // Size of hidden layer
		     size_t n_output, // Size of output layer
		     int a_add_bias, // Whether to add bias on this net
		     int a_in2hid_exp = 0, // Fixed point exponent of weights
		     int a_hid2out_exp = 0,
		     size_t size_bunch = 8 // bunch size
	            );
    ~QN_MLP_MultiSPERTFx3_slave();

    // Find out the size of the net
    size_t num_layers() const { return 3; };
    size_t num_sections() const { return 4; };
    size_t get_n_input() const { return n_input; };
    size_t get_n_output() const { return n_output; };
    size_t get_size_bunch() const { return size_bunch; };
    size_t size_layer(QN_LayerSelector layer) const;
    void size_section(QN_SectionSelector section, size_t* ouput_p,
		      size_t* input_p) const;

    size_t num_connections() const;
    void forward_bunch( size_t n_frames, const float* in, float* out);
    void back_bunch( size_t n_frames, const float* err);
    void fwd_bunchP( size_t n_frames, const fxint16* in, float* out);
    void bck_bunchP( size_t n_frames, const float* err);

    // Access weights.
    void set_weights(enum QN_SectionSelector which,
		     size_t row, size_t col,
		     size_t n_rows, size_t n_cols,
		     const float* weights);
    void get_weights(enum QN_SectionSelector which,
		     size_t row, size_t col,
		     size_t n_rows, size_t n_cols,
		     float* weights) const;

    // Access learning constants.
    void set_learnrate(enum QN_SectionSelector which, float rate);
    float get_learnrate(enum QN_SectionSelector which) const;
private:
    QN_ClassLogger clog;	// Logger object.

    const size_t n_input;	// No of input units
    const size_t n_hidden;	// No of hidden units
    const size_t n_output;	// No of output units
    const size_t n_in2hid;	// Number of in2hid weights.
    const size_t n_hid2out;	// Number of hid2out weights.

    int add_bias;		// 1 if we are to add in biases, else 0.

    // These exponenents are user settable
    const int in2hid_exp;
    const int hid2out_exp;

    const size_t size_bunch;

    fxint16 *in_y;		// Input as 16 bit fixed point.
    fxint16 *in_y_old;		// last Input as 16 bit fixed point.
    enum { in_y_exp = 3 };	// Exponent for above.
    fxint16 *in2hid;		// Input-to-hidden weights.
    fxint16 *in2hid_l;		// Input-to-hidden weights - low order bits.
    fxint32 *hid_bias;		// Hidden layer biases.
    enum { hid_bias_exp = 8 };  // was 10
    fxint32 *hid_x32;		// Sum into hidden layer.
    enum { hid_x32_exp = hid_bias_exp };
    fxint16 *hid_x16;		// Sum into hidden layer.
    enum { hid_x16_exp = 4 };
    fxint32 *bias_update;	// temp values during bias updates
    fxint16 *hid_y;		// Output from hidden units.
    fxint16 *hid_y_old;		// last Output from hidden units.
    enum { hid_y_exp = 0 };

    fxint16 *hid2out;		// Hidden-to-output weights.
    fxint16 *hid2out_l;		// Hidden-to-output weights  low order bits..
    fxint32 *out_bias;		// Output layer biases.
    enum { out_bias_exp = 8 }; // was 16
    fxint32 *out_x32;		// Sum into output layer.
    enum { out_x32_exp = out_bias_exp };  // was 10
    fxint16 *out_x16;		// Sum into output layer.
    enum { out_x16_exp = 4 };
    fxint16 *out_y;		// Output from output layer.
    enum { out_y_exp = 0 };

    //fxint16 *target16;		// Correct output
    //enum { target16_exp = out_y_exp };
    //fxint16 *out_dedy;		// Output error
    //enum { out_dedy_exp = 0 };
    //fxint16 *out_dydx;		// Output sigmoid difference
    //enum { out_dydx_exp = -2 };
    fxint16 *out_dedx;		// Error at output
    enum { out_dedx_exp = out_y_exp };
    fxint16 *hid2out_sdedx;		// Error scaled by learning rate
    // hid2out_sdedx_exp calculated in train1()
    
    fxint32 *hid_dedy32;	// Hidden output error.
    enum { hid_dedy32_exp = 6 };
    fxint16 *hid_dedy16;	// Hidden output error.
    enum { hid_dedy16_exp = 2 };
    fxint16 *hid_dydx;		// Hidden sigmoid difference
    enum { hid_dydx_exp = -2 };
    fxint16 *hid_dedx;		// Hidden feed back error term.
    enum { hid_dedx_exp = hid_dydx_exp + hid_dedy16_exp };
    fxint16 *in2hid_sdedx;		// Error scaled by learning rate
    // in2hid_sdedx_exp calculated in train1()
    // depends on in2hid_learnrate_exp

    fxint16 neg_in2hid_learnrate;	// Input to hidden learning rate
    fxexp_t neg_in2hid_learnrate_exp;	// Exponent for above
    fxint16* neg_hid_learnrate;		// Hidden bias learning rate
    fxexp_t neg_hid_learnrate_exp;	// Exponent for above
    fxint16 neg_hid2out_learnrate;	// Hidden to output learning rate 
    fxexp_t neg_hid2out_learnrate_exp;	// Exponent for above
    fxint16* neg_out_learnrate;		// Output bias learning rate
    fxexp_t neg_out_learnrate_exp;	// Exponent for above
    int feature_sat_warning_given;	// Used to suppress excess warning
					// messages.
    int weight_sat_warning_given;	// Used to suppress excess warning
					// messages.
    int bias_sat_warning_given;		// Used to suppress excess warning
					// messages.
    int lr_sat_warning_given;		// Used to suppress excess warning
  					// messages.
};

int
findexp_f_i(float f)
{
    return (int) ceil(log2(fabs(f)));
}


QN_MLP_MultiSPERTFx3_slave::QN_MLP_MultiSPERTFx3_slave(int dbg, size_t input, size_t hidden,
				   size_t output, int a_add_bias,
				   int a_in2hid_exp, int a_hid2out_exp,
				   size_t a_bunch_size )
    : clog(dbg, "QN_MLP_MultiSPERTFx3_slave", ""),
      n_input(input),
      n_hidden(hidden),
      n_output(output),
      n_in2hid(input * hidden),
      n_hid2out(hidden * output),
      add_bias(a_add_bias),
      in2hid_exp(a_in2hid_exp),
      hid2out_exp(a_hid2out_exp),
      size_bunch(a_bunch_size),
      feature_sat_warning_given(0),
      weight_sat_warning_given(0),
      bias_sat_warning_given(0),
      lr_sat_warning_given(0)
{
    const size_t size_hidden = n_hidden * size_bunch;
    const size_t size_input  = n_input  * size_bunch;
    const size_t size_output = n_output * size_bunch;

    in_y = malloc_vi16(size_input); copy_i16_vi16(size_input, 0, in_y);
    in_y_old = malloc_vi16(size_input); copy_i16_vi16(size_input, 0, in_y_old);
    in2hid = malloc_vi16(n_in2hid); copy_i16_vi16(n_in2hid, 0, in2hid);
    in2hid_l = malloc_vi16(n_in2hid); copy_i16_vi16(n_in2hid, 0, in2hid_l);
    hid_bias = malloc_vi32(n_hidden); copy_i32_vi32(n_hidden, 0, hid_bias);
    hid_x32 = malloc_vi32(size_hidden);
    hid_x16 = malloc_vi16(size_hidden);
    hid_y = malloc_vi16(size_hidden);
    hid_y_old = malloc_vi16(size_hidden);

    hid2out = malloc_vi16(n_hid2out); copy_i16_vi16(n_hid2out, 0, hid2out);
    hid2out_l = malloc_vi16(n_hid2out); copy_i16_vi16(n_hid2out, 0, hid2out_l);
    out_bias = malloc_vi32(n_output); copy_i32_vi32(n_output, 0, out_bias);
    out_x32 = malloc_vi32(size_output);
    out_x16 = malloc_vi16(size_output);
    out_y = malloc_vi16(size_output);

    out_dedx = malloc_vi16(size_output);
       copy_i16_vi16(size_output, 0, out_dedx);
    hid2out_sdedx = malloc_vi16(size_output);
    neg_hid_learnrate = malloc_vi16(size_bunch);
    neg_out_learnrate = malloc_vi16(size_bunch);
    bias_update = malloc_vi32((n_output>n_hidden)?n_output:n_hidden);

    hid_dedy32 = malloc_vi32(size_hidden);
    hid_dedy16 = malloc_vi16(size_hidden);
    hid_dydx = malloc_vi16(size_hidden);
    hid_dedx = malloc_vi16(size_hidden);
    in2hid_sdedx = malloc_vi16(size_hidden);
    clog.log(QN_LOG_PER_RUN, "Created net inputs=%u hidden=%u output=%u",
	     n_input, n_hidden, n_output);
    fprintf(stderr,"SPERT QN_MLP_MultiSPERTFx3 created\n");
}

QN_MLP_MultiSPERTFx3_slave::~QN_MLP_MultiSPERTFx3_slave()
{
    free_vi16(in_y);
    free_vi16(in2hid);
    free_vi32(hid_bias);
    free_vi32(hid_x32);
    free_vi16(hid_x16);
    free_vi16(hid_y);

    free_vi16(hid2out);
    free_vi32(out_bias);
    free_vi32(out_x32);
    free_vi16(out_x16);
    free_vi16(out_y);

    free_vi16(out_dedx);
    free_vi16(hid2out_sdedx);
    
    free_vi32(hid_dedy32);
    free_vi16(hid_dedy16);
    free_vi16(hid_dydx);
    free_vi16(hid_dedx);
    free_vi16(in2hid_sdedx);
}


size_t
QN_MLP_MultiSPERTFx3_slave::size_layer(QN_LayerSelector layer) const
{
    size_t r;

    switch(layer)
    {
    case QN_MLP3_INPUT:
	r = n_input;
	break;
    case QN_MLP3_HIDDEN:
	r = n_hidden;
	break;
    case QN_MLP3_OUTPUT:
	r = n_output;
	break;
    default:
	r = 0;
	break;
    }
    return r;
}

void
QN_MLP_MultiSPERTFx3_slave::size_section(QN_SectionSelector section, size_t* output_p,
			       size_t* input_p) const
{
    switch(section)
    {
    case QN_MLP3_INPUT2HIDDEN:
	*output_p = n_hidden;
	*input_p = n_input;
	break;
    case QN_MLP3_HIDDENBIAS:
	*output_p = n_hidden;
	*input_p = 1;
	break;
    case QN_MLP3_HIDDEN2OUTPUT:
	*output_p = n_output;
	*input_p = n_hidden;
	break;
    case QN_MLP3_OUTPUTBIAS:
	*output_p = n_output;
	*input_p = 1;
	break;
    default:
	clog.error("Tying to find size of unknown section.");
    }
}

size_t
QN_MLP_MultiSPERTFx3_slave::num_connections() const
{
    // Note how we take account of the biases separately
    return (n_input+n_output) * n_hidden + n_hidden + n_output;
}


void
QN_MLP_MultiSPERTFx3_slave::fwd_bunchP(size_t n_frames, const fxint16* in,
				  float* out)
{
    const size_t size_hidden = n_hidden * n_frames;
    const size_t size_input  = n_input  * n_frames;
    const size_t size_output = n_output * n_frames;

    fxbool sat;
    size_t i;

    // Swap old<->new inputs and hidden outputs to save old values
    fxint16* in_y_swap = in_y;
    in_y = in_y_old;
    in_y_old = in_y_swap;
    fxint16* hid_y_swap = hid_y;
    hid_y = hid_y_old;
    hid_y_old = hid_y_swap;

    //fprintf(stderr,"F%d\n",idx);

    // Don't Convert inputs
    // sat = conv_vf_vx16s(size_input, in_y_exp, in, in_y);
    //if (sat && !feature_sat_warning_given)
    //{
    //clog.warning("features saturated on input.");
    //feature_sat_warning_given = 1;
    //}
    copy_vi16_vi16(size_input, in, in_y);

    // First layer
#ifndef NBUNCH_ROUTINES
   sat = mul_srx_mi16mi16_i_mi32t(n_frames, n_input, n_hidden,
		   n_input, n_hidden, n_hidden,
		   in_y, in2hid,
		   hid_x32_exp-in_y_exp-in2hid_exp-1,
		   hid_x32);
    if (sat!=FXFALSE)
    {
       clog.error("Intermediate saturation in forward pass input to hidden "
		  "matrix multiply - check normalization and fixed point "
		  "exponent.");
    }
    fxint32* t_hid_x32 = hid_x32;
    for(i=0;i<n_frames;i++) {
       acc_vx32_vx32( n_hidden, hid_bias_exp, hid_x32_exp, hid_bias,t_hid_x32);
       t_hid_x32 += n_hidden;
    }
#else
    fxint16* t_in_y = in_y;
    fxint32* t_hid_x32 = hid_x32;
    for(i=0;i<n_frames;i++)
    {
      sat = mul_vx16mx16_vx32t ( n_input, n_hidden,
				 in_y_exp, in2hid_exp, hid_x32_exp,
				 t_in_y, in2hid, t_hid_x32);
      if (sat!=FXFALSE)
	{
	  clog.error("Intermediate saturation in forward pass input to hidden "
		     "matrix multiply - check normalization and fixed point "
		     "exponent.");
	}
      acc_vx32_vx32( n_hidden, hid_bias_exp, hid_x32_exp, hid_bias,t_hid_x32);
      t_in_y    += n_input;
      t_hid_x32 += n_hidden;
    }   
#endif
 
    conv_vx32_vx16(size_hidden, hid_x32_exp, hid_x16_exp, hid_x32, hid_x16);
    sigmoid_vx16_vx16(size_hidden, hid_x16_exp, hid_y_exp, hid_x16, hid_y);

    // Second layer
#ifndef NBUNCH_ROUTINES
    sat = mul_srx_mi16mi16_i_mi32t(n_frames, n_hidden, n_output,
		   n_hidden, n_output, n_output,
		   hid_y, hid2out, 
		   out_x32_exp-hid_y_exp-hid2out_exp-1,
		   out_x32);
    if (sat!=FXFALSE)
    {
       clog.error("Intermediate saturation in forward pass input to hidden "
		  "matrix multiply - check normalization and fixed point "
		  "exponent.");
    }
    if(add_bias)
    {
       fxint32* t_out_x32 = out_x32;
       for(i=0;i<n_frames;i++) {
	  acc_vx32_vx32( n_output, out_bias_exp, out_x32_exp,
			 out_bias, t_out_x32);
	  t_out_x32 += n_output;
       } 
    }   
#else
    fxint16* t_hid_y = hid_y;
    fxint32* t_out_x32 = out_x32;
    for(i=0;i<n_frames;i++)
    {
      sat = mul_vx16mx16_vx32t( n_hidden, n_output,
				hid_y_exp, hid2out_exp, out_x32_exp,
				t_hid_y, hid2out, t_out_x32);
      if (sat!=FXFALSE)
      {
	clog.error("Intermediate saturation in forward pass hidden to output "
		   "matrix multiply - check normalization and fixed point "
		   "exponent.");
      }

      if(add_bias)
      {
	acc_vx32_vx32( n_output, out_bias_exp, out_x32_exp,
		       out_bias, t_out_x32);

      } 
      t_hid_y += n_hidden;
      t_out_x32 += n_output;
    }   
#endif
    conv_vx32_vf(size_output, out_x32_exp, out_x32, out);
}

void
QN_MLP_MultiSPERTFx3_slave::bck_bunchP(size_t n_frames, const float* err)
{
   if(n_frames>size_bunch)   // a hack to indicate final update
   {
      n_frames-=size_bunch;
      // Swap old<->new inputs and hidden outputs to get latest values
      fxint16* in_y_swap = in_y;
      in_y = in_y_old; 
      in_y_old = in_y_swap;
      fxint16* hid_y_swap = hid_y;
      hid_y = hid_y_old;
      hid_y_old = hid_y_swap;
   }
    const size_t size_hidden = n_hidden * n_frames;
    const size_t size_output = n_output * n_frames;
    fxbool sat;

// error was computed by host: just convert
    sat = conv_vf_vx16s(size_output, out_dedx_exp, err, out_dedx);

    if (sat!=FXFALSE)
    {
      for(size_t g=0;g<size_output;g++)
	if(err[g]>0.99)
	  printf("err[%d]= %f\n",g,err[g]);
      printf("\n");
      fflush(stdout) ;
      clog.warning("saturation in converting error from float in backP");
    }


// Back propagate error through hidden-to-output weights
// note: Incorrectly uses latest weight matrix for pipeline mode
    sat = mulnt_srx_mi16mi16_i_mi32t(n_frames, n_output, n_hidden,
				   n_output, n_output, n_hidden,
				   out_dedx, hid2out,
				   -out_dedx_exp-hid2out_exp+hid_dedy32_exp-1,
				   hid_dedy32);
    if (sat!=FXFALSE)
      {
	clog.error("Intermediate saturation in training hidden to output "
		   "back propagation - check error terms and fixed point "
		   "exponent.");
      }
    conv_vx32_vx16( size_hidden, hid_dedy32_exp, hid_dedy16_exp,
		    hid_dedy32, hid_dedy16);

// Propagate error back through sigmoid with old values of hid_y (pipe!)
    dsigmoid_vx16_vx16( size_hidden, hid_y_exp, hid_dydx_exp,
		        hid_y_old, hid_dydx);
    mul_vx16vx16_vx16(size_hidden, hid_dydx_exp, hid_dedy16_exp, hid_dedx_exp,
		      hid_dydx, hid_dedy16, hid_dedx);

// Do loop invariants first
  // scale out_dedx
    fxexp_t hid2out_sdedx_exp = out_y_exp + neg_hid2out_learnrate_exp;
    mul_vx16x16_vx16(size_output,
		     out_dedx_exp, neg_hid2out_learnrate_exp,hid2out_sdedx_exp,
		     out_dedx, neg_hid2out_learnrate,hid2out_sdedx);
  // scale hid_dedx
    fxexp_t in2hid_sdedx_exp = hid_dedx_exp + neg_in2hid_learnrate_exp;
    mul_vx16x16_vx16(size_hidden,
		     hid_dedx_exp, neg_in2hid_learnrate_exp, in2hid_sdedx_exp,
		     hid_dedx, neg_in2hid_learnrate, in2hid_sdedx);

// Update hid and out biases (out biases only in one partial net)
    if(add_bias)
      {
	sat = mul_vx16mx16_vx32t( n_frames, n_output,
			      neg_out_learnrate_exp,out_dedx_exp,out_bias_exp,
			      neg_out_learnrate, out_dedx, bias_update); 
	satadd_vi32vi32_vi32( n_output, bias_update, out_bias, out_bias);
      }

    sat = mul_vx16mx16_vx32t( n_frames, n_hidden,
			      neg_hid_learnrate_exp,hid_dedx_exp, hid_bias_exp,
			      neg_hid_learnrate, hid_dedx, bias_update);
    satadd_vi32vi32_vi32( n_hidden, bias_update, hid_bias, hid_bias); 

// Update hid2out weights (32 bits!)
    sat = multn_srx_srx_satadd_mi16mi16_i_i_mhi32_mhi32st(
				 n_hidden, n_frames, n_output,
				 n_hidden, n_output, n_output, n_output,
				 hid_y_old, hid2out_sdedx,
				 -hid_y_exp-hid2out_sdedx_exp+hid2out_exp-1,0,
				 hid2out, hid2out_l, hid2out, hid2out_l);
// Update in2hid weights and biases
    sat = multn_srx_srx_satadd_mi16mi16_i_i_mhi32_mhi32st(
				 n_input, n_frames, n_hidden,
				 n_input, n_hidden, n_hidden, n_hidden,
				 in_y_old, in2hid_sdedx,
				 -in_y_exp-in2hid_sdedx_exp+in2hid_exp-1,0,
				 in2hid, in2hid_l, in2hid, in2hid_l);
}

void
QN_MLP_MultiSPERTFx3_slave::set_weights(enum QN_SectionSelector which, 
			      size_t row, size_t col, 
			      size_t n_rows, size_t n_cols,
			      const float* weights)
{
    fxbool weight_sat = FXFALSE;
    fxbool bias_sat = FXFALSE;

    // Note that the fixed point weight matrix is transposed
    switch(which)
    {
    case QN_MLP3_INPUT2HIDDEN:
    {
	const float* src_ptr = weights;
	fxint16* dest_ptr = in2hid + col*n_hidden + row;
	fxint16* buf = falloc_vi16(n_cols);
	for (size_t i=0; i<n_rows; i++)
	{
	    weight_sat |= conv_vf_vx16s(n_cols, in2hid_exp, src_ptr, buf);
	    copy_vi16_svi16(n_cols, n_hidden, buf, dest_ptr);
	    src_ptr += n_cols;
	    dest_ptr += 1;
	}
	ffree_vi16(buf);
	break;
    }
    case QN_MLP3_HIDDEN2OUTPUT:
    {
	const float* src_ptr = weights;
	fxint16* dest_ptr = hid2out + col*n_output + row;
	fxint16* buf = falloc_vi16(n_cols);
	for (size_t i=0; i<n_rows; i++)
	{
	    weight_sat |= conv_vf_vx16s(n_cols, hid2out_exp, src_ptr, buf);
	    copy_vi16_svi16(n_cols, n_output, buf, dest_ptr);
	    src_ptr += n_cols;
	    dest_ptr += 1;
	}
	ffree_vi16(buf);
	break;
    }
    case QN_MLP3_HIDDENBIAS:
	bias_sat |= 
	    conv_vf_vx32s(n_rows, hid_bias_exp, weights, hid_bias + row);
	break;
    case QN_MLP3_OUTPUTBIAS:
	bias_sat |=
	    conv_vf_vx32s(n_rows, out_bias_exp, weights, out_bias + row);
	break;
    default:
	clog.error("trying to set weights in unknown section %d.",which);
    }
    // This next bit gives us at most one warning of saturated weights/biases
    if (weight_sat && !weight_sat_warning_given)
    {
	clog.warning("weights saturated on input.");
	weight_sat_warning_given = 1;
    }
    if (bias_sat && !bias_sat_warning_given)
    {
	clog.warning("biases saturated on input.");
	bias_sat_warning_given = 1;
    }
}

void 
QN_MLP_MultiSPERTFx3_slave::get_weights(enum QN_SectionSelector which,
			      size_t row, size_t col,
			      size_t n_rows, size_t n_cols,
			      float* weights) const
{
    // Note that the fixed point weight matrix is transposed
    switch(which)
    {
    case QN_MLP3_INPUT2HIDDEN:
    {
	float* dest_ptr = weights;
	fxint16* src_ptr = in2hid + col*n_hidden + row;
	fxint16* buf = falloc_vi16(n_cols);
	for (size_t i=0; i<n_rows; i++)
	{
	    copy_svi16_vi16(n_cols, n_hidden, src_ptr, buf);
	    conv_vx16_vf(n_cols, in2hid_exp, buf, dest_ptr);
	    src_ptr += 1;
	    dest_ptr += n_cols;
	}
	ffree_vi16(buf);
	break;
    }
    case QN_MLP3_HIDDEN2OUTPUT:
    {
	float* dest_ptr = weights;
	fxint16* src_ptr = hid2out + col*n_output + row;
	fxint16* buf = falloc_vi16(n_cols);
	for (size_t i=0; i<n_rows; i++)
	{
	    copy_svi16_vi16(n_cols, n_output, src_ptr, buf);
	    conv_vx16_vf(n_cols, hid2out_exp, buf, dest_ptr);
	    src_ptr += 1;
	    dest_ptr += n_cols;
	}
	ffree_vi16(buf);
	break;
    }
    case QN_MLP3_HIDDENBIAS:
	conv_vx32_vf(n_rows, hid_bias_exp, hid_bias + row, weights);
	break;
    case QN_MLP3_OUTPUTBIAS:
	conv_vx32_vf(n_rows, out_bias_exp, out_bias + row, weights);
	break;
    default:
	clog.error("trying to get weights from unknown section.");
    }
}

void
QN_MLP_MultiSPERTFx3_slave::set_learnrate(enum QN_SectionSelector which, float learnrate)
{
    fxbool sat = FXFALSE;
    enum { MIN_LEARNRATE_EXP = -10 };

    switch(which)
    {
    case QN_MLP3_INPUT2HIDDEN:
	neg_in2hid_learnrate_exp = findexp_f_i(-learnrate);
	if (neg_in2hid_learnrate_exp < MIN_LEARNRATE_EXP)
	    neg_in2hid_learnrate_exp = MIN_LEARNRATE_EXP;
	sat = conv_f_x16s(neg_in2hid_learnrate_exp, -learnrate,
			  &neg_in2hid_learnrate);
	clog.log(QN_LOG_PER_EPOCH,  "Set in2hid learning rate %f (exp %i).",
		   -conv_x16_f(neg_in2hid_learnrate_exp,neg_in2hid_learnrate),
		   neg_in2hid_learnrate_exp);
	break;
    case QN_MLP3_HIDDEN2OUTPUT:
	neg_hid2out_learnrate_exp = findexp_f_i(-learnrate);
	if (neg_hid2out_learnrate_exp < MIN_LEARNRATE_EXP)
	    neg_hid2out_learnrate_exp = MIN_LEARNRATE_EXP;
	sat = conv_f_x16s(neg_hid2out_learnrate_exp, -learnrate,
			  &neg_hid2out_learnrate);
  	clog.log(QN_LOG_PER_EPOCH, "Set hid2out learning rate %f (exp %i).",
		 -conv_x16_f(neg_hid2out_learnrate_exp,
			     neg_hid2out_learnrate),
		 neg_hid2out_learnrate_exp);
	break;
    case QN_MLP3_HIDDENBIAS:
        fxint16 lneg_hid_learnrate;
	neg_hid_learnrate_exp = findexp_f_i(-learnrate);
	if (neg_hid_learnrate_exp < MIN_LEARNRATE_EXP)
	    neg_hid_learnrate_exp = MIN_LEARNRATE_EXP;
	sat = conv_f_x16s(neg_hid_learnrate_exp, -learnrate,
			  &lneg_hid_learnrate);
	for(size_t i=0;i<size_bunch;i++)
	  neg_hid_learnrate[i]=lneg_hid_learnrate;
	clog.log(QN_LOG_PER_EPOCH,"Set hidbias learning rate %f (exp %i).",
		 -conv_x16_f(neg_hid_learnrate_exp, lneg_hid_learnrate),
		 neg_hid_learnrate_exp);
	break;
    case QN_MLP3_OUTPUTBIAS:
        fxint16 lneg_out_learnrate;
	neg_out_learnrate_exp = findexp_f_i(-learnrate);
	if (neg_out_learnrate_exp < MIN_LEARNRATE_EXP)
	    neg_out_learnrate_exp = MIN_LEARNRATE_EXP;
	sat = conv_f_x16s(neg_out_learnrate_exp, -learnrate,
			  &lneg_out_learnrate);
	for(size_t i=0;i<size_bunch;i++)
	  neg_out_learnrate[i]=lneg_out_learnrate;
	clog.log(QN_LOG_PER_EPOCH, "Set outbias learning rate %f (exp %i).",
		 -conv_x16_f(neg_out_learnrate_exp, lneg_out_learnrate),
		 neg_out_learnrate_exp);
	break;
    default:
	clog.error("trying to set learning rate in unknown section.");
	break;
    }
    // This next bit gives us at most on warning of saturated learning rate
    // Note - currently this does nothing - but no harm in leaving it
    // in case we change the way learning rates work again - davidj 9/13/95
    if (sat && !lr_sat_warning_given)
    {
	clog.warning("learning rate saturated.");
	lr_sat_warning_given = 1;
    }
}

float 
QN_MLP_MultiSPERTFx3_slave::get_learnrate(enum QN_SectionSelector which) const
{
    float res;			// Returned learning rate

    switch(which)
    {
    case QN_MLP3_INPUT2HIDDEN:
	res = -conv_x16_f(neg_in2hid_learnrate_exp, neg_in2hid_learnrate);
	break;
    case QN_MLP3_HIDDEN2OUTPUT:
	res = -conv_x16_f(neg_hid2out_learnrate_exp, neg_hid2out_learnrate);
	break;
    case QN_MLP3_HIDDENBIAS:
	res = -conv_x16_f(neg_hid_learnrate_exp, neg_hid_learnrate[0]);
	break;
    case QN_MLP3_OUTPUTBIAS:
	res = -conv_x16_f(neg_out_learnrate_exp, neg_out_learnrate[0]);
	break;
    default:
	res = 0;
	clog.error("trying to get learning rate from unknown section.");
	break;
    }
    return res;
}



#define BUFFERSIZE 1900000

/* Only one Global MLP on Spert */
QN_MLP_MultiSPERTFx3_slave *GlobalMLP;

void do_constructor( SPC_ARGS args )
{
  int        debug       = *spc_getarg_i( args );
  size_t     n_input     = *spc_getarg_u( args );
  size_t     n_hidden    = *spc_getarg_u( args );
  size_t     n_output    = *spc_getarg_u( args );
  int        add_bias    = *spc_getarg_u( args );
  int        in2hid_exp  = *spc_getarg_i( args );
  int        hid2out_exp = *spc_getarg_i( args );
  size_t     bunch_size  = *spc_getarg_u( args );

  GlobalMLP = new QN_MLP_MultiSPERTFx3_slave(
                        debug, n_input, n_hidden, n_output,
			add_bias, in2hid_exp, hid2out_exp, bunch_size );
}


void do_destructor( SPC_ARGS args )
{
  *args=0;
  delete GlobalMLP;
  exit(0);
  /* never gets here */
}

void do_fwd_bunchP( SPC_ARGS args )
{
  size_t     n_frames = *spc_getarg_u( args );
  fxint16*   input    = spc_getarg_vi16( args );
  float*     output   = spc_getarg_vf( args );

  GlobalMLP->fwd_bunchP( n_frames, input, output );
}

void do_bck_bunchP( SPC_ARGS args )
{
  size_t   n_frames = *spc_getarg_u( args );
  float*   error    = spc_getarg_vf( args );

  GlobalMLP->bck_bunchP( n_frames, error );
}

void do_set_weights( SPC_ARGS args )
{
  QN_SectionSelector which = *(QN_SectionSelector*)spc_getarg_u( args );
  size_t   row     = *spc_getarg_u( args );
  size_t   col     = *spc_getarg_u( args );
  size_t   n_rows  = *spc_getarg_u( args );
  size_t   n_cols  = *spc_getarg_u( args );
  float*   weights = spc_getarg_vf( args );
  
  GlobalMLP->set_weights( which, row, col, n_rows, n_cols, weights);
}

void do_get_weights( SPC_ARGS args )
{
  QN_SectionSelector which = *(QN_SectionSelector*)spc_getarg_u( args );
  size_t   row     = *spc_getarg_u( args );
  size_t   col     = *spc_getarg_u( args );
  size_t   n_rows  = *spc_getarg_u( args );
  size_t   n_cols  = *spc_getarg_u( args );
  float*   weights = spc_getarg_vf( args );

  GlobalMLP->get_weights( which, row, col, n_rows, n_cols, weights);
}

void do_get_learnrate( SPC_ARGS args )
{
  QN_SectionSelector which = *(QN_SectionSelector*)spc_getarg_u( args );
  float* result = spc_getarg_vf( args );

  *result = GlobalMLP->get_learnrate( which );
}

void do_set_learnrate( SPC_ARGS args )
{
  QN_SectionSelector which = *(QN_SectionSelector*)spc_getarg_u( args );
  float      learnrate     = *spc_getarg_f( args );

  GlobalMLP->set_learnrate( which, learnrate);

}

main()
{
    SPC_CALL CallTbl[20];

    CallTbl[ QN_SPC_MultiSPERTFx3::CONSTRUCT ] = do_constructor;
    CallTbl[ QN_SPC_MultiSPERTFx3::DESTRUCT ] = do_destructor;
    CallTbl[ QN_SPC_MultiSPERTFx3::SETWEIGHTS ] = do_set_weights;
    CallTbl[ QN_SPC_MultiSPERTFx3::GETWEIGHTS ] = do_get_weights;
    CallTbl[ QN_SPC_MultiSPERTFx3::SETLEARN ] = do_set_learnrate;
    CallTbl[ QN_SPC_MultiSPERTFx3::GETLEARN ] = do_get_learnrate;
    CallTbl[ QN_SPC_MultiSPERTFx3::BACKWARD ] = do_bck_bunchP;
    CallTbl[ QN_SPC_MultiSPERTFx3::FORWARD ] = do_fwd_bunchP;
 
    SPC_EVENT_LOOP( CallTbl , BUFFERSIZE );
}


#else
// Dummy main in case this gets compiled on the host.
int
main(int, const char*)
{
    fprintf(stderr, "qnspc_multispertfx3_1: ERROR - this program "
	    "does nothing on this system\n.");
    exit(EXIT_SUCCESS);
}
#endif // #if defined(HAVE_LIBSPCSLAVE) && defined(HAVE_LIBFXLIB)
