#ifndef NO_RCSID
const char* qntrain_rcsid = 
    "$Header: /n/lasagna/db/davidj/spert/src/quicknet/RCS/qntrain.cc,v 1.75 1998/06/12 01:25:29 davidj Exp $";
#endif

#include <QN_config.h>
#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef QN_HAVE_LIMITS_H
#include <limits.h>
#endif
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS (0)
#define EXIT_FAILURE (1)
#endif
#include <sys/types.h>
#ifdef QN_HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef QN_HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#include <unistd.h>

extern "C" void srand48(long);
#ifdef QN_HAVE_GETHOSTNAME
extern "C" int gethostname(...);
#endif
#ifdef __GNUC__
extern "C" {
    typedef void (*new_handler)(void);
    new_handler set_new_handler (new_handler);
}
#endif

#ifndef FILENAME_MAX
#define FILENAME_MAX (MAXPATHLEN)
#endif

#include "fltvec.h"
#include "QuickNet.h"

// Some distinctive default values for disused options.
#define DEFAULT_INITIAL_LEARNRATE (0.0f)
#define DEFAULT_TRAIN_CACHE_SENTS (0)

static struct {
    char* train_ftrfile_filename;
    char* train_ftrfile_format;
    int train_ftrfile_num_ftrs;
    char* inweights_filename;
    char* outweights_filename;
    char* logweights_filename;
    char* innorms_filename;
    float min_random_bias;
    float max_random_bias;
    float min_random_weight;
    float max_random_weight;
    // This is here for historical compatability.
    float initial_learnrate;
    char* learnrate_schedule;
    QN_Arg_ListFloat learnrate;
    float learnrate_scaleby;
    float learnrate_ramp_thresh;
    float learnrate_stop_thresh;
    long max_epochs;
    int window_width;
    int window_offset;
    int num_input;
    int num_hidden;
    int num_output;
    char* mlp_output_type;
    int bunch_size;
    long first_cv_sent;
    long num_cv_sents;
    long first_train_sent;
    long num_train_sents;
    long train_cache_sents;
    long train_cache_frames;
    int first_ftr;
    int num_ftrs;
    int use_fixed;
    int seed;
    int verbose;
    int dbg;			// Debug level.
} config;

static void
set_defaults(void)
{
    static float default_learnrate[1] = { 0.008 };

    config.train_ftrfile_filename = "(unset)";
    config.train_ftrfile_format = "pfile";
    config.train_ftrfile_num_ftrs = 0;
    config.inweights_filename = "-";
    config.outweights_filename = "out.weights";
    config.logweights_filename = "log.weights";
    config.innorms_filename = "";
    config.min_random_bias = -4.1;
    config.max_random_bias = -3.9;
    config.min_random_weight = -0.1;
    config.max_random_weight = 0.1;
    config.initial_learnrate = DEFAULT_INITIAL_LEARNRATE;
    config.learnrate_schedule = "newbob";
    config.learnrate.count = 1;
    config.learnrate.vals = &default_learnrate[0];
    config.learnrate_scaleby = 0.5;
    config.learnrate_stop_thresh = 0.5;
    config.learnrate_ramp_thresh = 0.5;
    config.max_epochs = LONG_MAX;
    config.window_width = 9;
    config.window_offset = -4;
    config.num_input = 153;
    config.num_hidden = 200;
    config.num_output = 56;
    config.mlp_output_type = "softmax";
    config.bunch_size = 1;
    config.first_train_sent = 0;
    config.num_train_sents = 1720;
    config.first_cv_sent = 1720;
    config.num_cv_sents = 230;
    config.train_cache_sents = DEFAULT_TRAIN_CACHE_SENTS;
    config.train_cache_frames = 10000;
    config.first_ftr = 1;
    config.num_ftrs = 0;
    config.use_fixed = 0;
    config.seed = 0;
    config.verbose = 1;
    config.dbg = 0;
}

QN_ArgEntry argtab[] =
{
{ NULL, "Fast phoneme recognition MLP trainer version " QN_VERSION, QN_ARG_DESC },
{ "train_ftrfile", "Input feature file", QN_ARG_STR,
  &(config.train_ftrfile_filename), QN_ARG_REQ },
{ "train_ftrfile_format", "Feature file format [pfile,pre]", QN_ARG_STR,
  &(config.train_ftrfile_format) },
{ "train_ftrfile_num_ftrs", "Width of feature file", QN_ARG_INT,
  &(config.train_ftrfile_num_ftrs) },
{ "input_normfile", "Input normalization file", QN_ARG_STR,
  &(config.innorms_filename) },
{ "input_weightfile", "Input weight file", QN_ARG_STR,
  &(config.inweights_filename) },
{ "output_weightfile", "Output weight file", QN_ARG_STR,
  &(config.outweights_filename) },
{ "log_weightfile", "Log weight file", QN_ARG_STR,
  &(config.logweights_filename) },
{ "min_random_bias", "Minimum random bias", QN_ARG_FLOAT,
  &(config.min_random_bias) },
{ "max_random_bias", "Maximum random bias", QN_ARG_FLOAT,
  &(config.max_random_bias) },
{ "min_random_weight", "Minimum random weight", QN_ARG_FLOAT,
  &(config.min_random_weight) },
{ "max_random_weight", "Maximum random weight", QN_ARG_FLOAT,
  &(config.max_random_weight) },
{ "initial_learnrate", "For historical compatability", QN_ARG_FLOAT,
  &(config.initial_learnrate) },
{ "learnrate_schedule", "LR schedule type [newbob,list]",
      QN_ARG_STR, &(config.learnrate_schedule) },
{ "learnrate", "Learning rates",
      QN_ARG_LIST_FLOAT, &(config.learnrate) },
{ "learnrate_scaleby", "Learning rate ramp scaling factor",
      QN_ARG_FLOAT, &(config.learnrate_scaleby) },
{ "learnrate_ramp_thresh", "Min delta-error to start ramping",
      QN_ARG_FLOAT, &(config.learnrate_ramp_thresh) },
{ "learnrate_stop_thresh", "Min delta-error to stop",
      QN_ARG_FLOAT, &(config.learnrate_stop_thresh) },
{ "max_epochs", "Maximum number of epochs", QN_ARG_LONG,
  &(config.max_epochs) },
{ "window_width", "Number if frames in input window", QN_ARG_INT,
  &(config.window_width) },
{ "window_offset", "Offset of input window", QN_ARG_INT,
  &(config.window_offset) },
{ "num_input", "Number of units in input layer",
  QN_ARG_INT, &(config.num_input)},
{ "num_hidden","Number of units in hidden layer",
  QN_ARG_INT, &(config.num_hidden) },
{ "num_output","Number of units in output layer",
  QN_ARG_INT, &(config.num_output) },
{ "mlp_output_type","Type of non-linearity in MLP output layer",
  QN_ARG_STR, &(config.mlp_output_type) },
{ "bunch_size","No. of frames per MLP forward/backward pass",
  QN_ARG_INT, &(config.bunch_size) },
{ "first_train_sent", "Number of first training sentence",
  QN_ARG_LONG, &(config.first_train_sent) },
{ "num_train_sents", "Number of training sentences",
  QN_ARG_LONG, &(config.num_train_sents) },
{ "first_cv_sent", "Number of first cross validation sentence",
  QN_ARG_LONG, &(config.first_cv_sent) },
{ "num_cv_sents", "Number of cross validation sentences",
  QN_ARG_LONG, &(config.num_cv_sents) },
{ "train_cache_sents", "Number of training sentences in cache",
  QN_ARG_LONG, &(config.train_cache_sents) },
{ "train_cache_frames", "Number of training frames in cache",
  QN_ARG_LONG, &(config.train_cache_frames) },
{ "first_ftr", "First feature used from feature database",
  QN_ARG_INT, &(config.first_ftr) },
{ "num_ftrs", "Number of features used from feature database",
  QN_ARG_INT, &(config.num_ftrs) },
{ "use_fx","Use fixed point",
  QN_ARG_BOOL, &(config.use_fixed) },
{ "seed", "Random number seed",
  QN_ARG_INT, &(config.seed) },
{ "verbose", "Output extra status messages",
  QN_ARG_BOOL, &(config.verbose) },
{ "debug", "Output internal diagnostic information",
  QN_ARG_INT, &(config.dbg) },
{ NULL, NULL, QN_ARG_NOMOREARGS }
};


static void
train_run(FILE* ftrfile, QN_StreamType ftrfile_type,
	  FILE* inweights, FILE* norms_file, FILE* outweights,
	  int verbose)
{
    int window_width = config.window_width;
    int window_offset = config.window_offset;

    size_t n_input = config.num_input;
    size_t n_hidden = config.num_hidden;
    size_t n_output = config.num_output;

    size_t first_train_sent = config.first_train_sent;
    size_t n_train_sent = config.num_train_sents;
    if (n_train_sent==INT_MAX)
	n_train_sent = QN_ALL;
    size_t first_cv_sent = config.first_cv_sent;
    size_t n_cv_sent = config.num_cv_sents;
    if (n_cv_sent==INT_MAX)
	n_cv_sent = QN_ALL;
    const char* logweights_filename = config.logweights_filename;

    int use_fixed = config.use_fixed;
    time_t now;

// A note for the logfile.
    time(&now);
    QN_OUTPUT("Program version: qntrain %s.", QN_VERSION);
#ifdef QN_HAVE_GETHOSTNAME
    char hostname[MAXHOSTNAMELEN];
    assert(gethostname(hostname, MAXHOSTNAMELEN)==0);
    QN_OUTPUT("Host: %s.", hostname);
#endif
    QN_OUTPUT("Program start: %.24s.", ctime(&now));

// Work out the learning rate stuff early to squash mistakes ASAP.
    QN_RateSchedule* rate_sched;
    if (strcmp(config.learnrate_schedule, "newbob")==0)
    {
	rate_sched = new QN_RateSchedule_NewBoB(config.learnrate.vals[0],
						config.learnrate_scaleby,
						config.learnrate_ramp_thresh,
						config.learnrate_stop_thresh,
						100.0f, 
						config.max_epochs);
    }
    else if (strcmp(config.learnrate_schedule, "list")==0)
    {
	long count;

	if (config.max_epochs < config.learnrate.count)
	    count = config.max_epochs;
	else
	    count = config.learnrate.count;
	rate_sched = new QN_RateSchedule_List(config.learnrate.vals, count);
    }
    else
    {
	QN_ERROR(NULL, "Unknown learning rate schedule '%s'.",
		 config.learnrate_schedule);
	rate_sched = NULL;
    }
    

// The basic database we will use.
    if (verbose>0)
    {
	QN_OUTPUT("Opening feature file...");
    }
    QN_InFtrLabStream* ftrlab_str; // The feature and label input stream.
    switch(ftrfile_type)
    {
    case QN_STREAM_PFILE:
    {
	QN_InFtrLabStream_PFile* pfile_str =
	    new QN_InFtrLabStream_PFile(config.dbg, // Select debugging.
					"input", // Debugging tag.
					ftrfile, // Feature file.
					1 // Indexed flag.
		);
	if (config.train_ftrfile_num_ftrs!=0
	    && pfile_str->num_ftrs()!=(size_t) config.train_ftrfile_num_ftrs)
	{
	    QN_ERROR(NULL, "Feature file has %lu features, command line"
		     " specified %lu features.",
		     (unsigned long) pfile_str->num_ftrs(),
		     config.train_ftrfile_num_ftrs);
	}
	ftrlab_str = pfile_str;
	break;
    }
    case QN_STREAM_PREFILE:
    {
	QN_InFtrLabStream_PreFile* prefile_str =
	    new QN_InFtrLabStream_PreFile(config.dbg, // Select debugging.
					  "input", // Debugging tag.
					  ftrfile,	// Feature file.
					  config.train_ftrfile_num_ftrs,
					  1 // Indexed
		);
	ftrlab_str = prefile_str;
	break;
    }
    default:
	ftrlab_str = NULL;
	assert(0);
    }

    size_t ftrlab_num_sents = ftrlab_str->num_segs();
    size_t ftrlab_num_frames = ftrlab_str->num_frames();
    size_t ftrlab_num_ftrs = ftrlab_str->num_ftrs();
    QN_OUTPUT("Feature file opened - %lu sentences, %lu frames, "
	      "%lu features in each frame.",
	      (unsigned long) ftrlab_num_sents,
	      (unsigned long) ftrlab_num_frames,
	      (unsigned long) ftrlab_num_ftrs);

    // Split the combined feature and label stream into separate,
    // asynchronous streams.
    QN_InFtrStream_SplitFtrLab* split_ftr_str =
	new QN_InFtrStream_SplitFtrLab(config.dbg, "split_ftr",
				       *ftrlab_str, 500);
    QN_InLabStream_SplitFtrLab* split_lab_str =
	new QN_InLabStream_SplitFtrLab(*split_ftr_str);
    QN_InFtrStream* ftr_str = split_ftr_str;
    QN_InLabStream* lab_str = split_lab_str;
    size_t num_ftrs = ftr_str->num_ftrs();
    size_t num_labs = lab_str->num_labs();
    assert(num_labs==1);

    // If a normalization file is used, normalize the feature stream.
    QN_InFtrStream_Norm* norm_str = NULL;
    if (norms_file!=NULL)
    {
	float* norm_mean = new float[num_ftrs];
	float* norm_sdevs = new float[num_ftrs];
	float* norm_bias = new float[num_ftrs];
	float* norm_scale = new float[num_ftrs];
	if (QN_read_norms_rap(norms_file, num_ftrs,
			      norm_mean, norm_sdevs))
	{
	    QN_ERROR(NULL, "failed to read norm file.");
	}
	// We actually want to subtract the mean and divide by the standard
	// deviation - the "Norm" filter adds a bias and multiplies by a
	// scale.
	neg_vf_vf(num_ftrs, norm_mean, norm_bias);
	recip_vf_vf(num_ftrs, norm_sdevs, norm_scale);
	norm_str = new QN_InFtrStream_Norm(config.dbg, "norm", *ftr_str,
					   norm_bias, norm_scale);
	ftr_str = norm_str;
	delete norm_scale;
	delete norm_bias;
	delete norm_sdevs;
	delete norm_mean;
    }

    // If we do not want all the features, extract just the ones we need.
    QN_InFtrStream_Narrow* narrow_str = NULL;
    if (config.first_ftr!=0
	|| ((size_t) config.num_ftrs!=num_ftrs && config.num_ftrs!=0)
	)
    {
	size_t narrow_num_ftrs;
	if (config.num_ftrs==0)
	    narrow_num_ftrs = QN_ALL;
	else
	    narrow_num_ftrs = config.num_ftrs;
	narrow_str = new QN_InFtrStream_Narrow(config.dbg, "narrow",
					       *ftr_str, config.first_ftr,
					       narrow_num_ftrs, 500);
	ftr_str = narrow_str;
    }

    // Create training and cross-validation streams.
    QN_InFtrStream_Cut* train_ftr_str = NULL;
    QN_InLabStream_Cut* train_lab_str = NULL;
    QN_InFtrStream_Cut2* cv_ftr_str = NULL;
    QN_InLabStream_Cut2* cv_lab_str = NULL;

    train_ftr_str = new QN_InFtrStream_Cut(config.dbg, "trainftr", *ftr_str,
					   first_train_sent, n_train_sent,
					   first_cv_sent, n_cv_sent);
    cv_ftr_str = new QN_InFtrStream_Cut2(*train_ftr_str);
    train_lab_str = new QN_InLabStream_Cut(config.dbg, "trainlab", *lab_str,
					   first_train_sent, n_train_sent,
					   first_cv_sent, n_cv_sent);
    cv_lab_str = new QN_InLabStream_Cut2(*train_lab_str);

// Window the streams for training.
    if (window_offset>0 || window_offset<(-window_width+1))
	QN_ERROR(NULL, "Label position must be within feature window.");
    size_t lab_win_len = 1;
    size_t lab_top_margin = (size_t) -window_offset;
    size_t lab_bot_margin = window_width + window_offset - lab_win_len;
    size_t ftr_win_len = window_width;
    size_t ftr_top_margin = 0;
    size_t ftr_bot_margin = 0;

    QN_InFtrStream_RandWindow* train_winftr_str =
	new QN_InFtrStream_RandWindow(config.dbg, "train_ftrwin",
				      *train_ftr_str, ftr_win_len,
				      ftr_top_margin, ftr_bot_margin,
				      config.train_cache_frames, config.seed
				      );
    QN_InLabStream_RandWindow* train_winlab_str =
	new QN_InLabStream_RandWindow(config.dbg, "train_labwin",
				      *train_lab_str, lab_win_len,
				      lab_top_margin, lab_bot_margin,
				      config.train_cache_frames, config.seed
				      );
    QN_InFtrStream_SeqWindow* cv_winftr_str =
	new QN_InFtrStream_SeqWindow(config.dbg, "cv_ftrwin",
				     *cv_ftr_str, ftr_win_len,
				     ftr_top_margin, ftr_bot_margin
				     );
    QN_InLabStream_SeqWindow* cv_winlab_str =
	new QN_InLabStream_SeqWindow(config.dbg, "cv_labwin",
				     *cv_lab_str, lab_win_len,
				     lab_top_margin, lab_bot_margin
				     );

    // Create MLP and load weights.
    QN_MLP* mlp3;

    QN_OutputLayerType outlayer_type;
    if (strcmp(config.mlp_output_type, "sigmoid")==0)
	outlayer_type = QN_OUTPUT_SIGMOID;
    else if (strcmp(config.mlp_output_type, "sigmoidx")==0)
	outlayer_type = QN_OUTPUT_SIGMOID_XENTROPY;
    else if (strcmp(config.mlp_output_type, "softmax")==0)
	outlayer_type = QN_OUTPUT_SOFTMAX;
    else
    {
	QN_ERROR(NULL, "unknown output unit type '%s'.",
		 config.mlp_output_type);
	outlayer_type = QN_OUTPUT_SIGMOID;
    }
    if (use_fixed)
    {
#ifdef QN_HAVE_LIBFXLIB
	if (config.bunch_size==1)
	{
	    mlp3 = new QN_MLP_Online1632Fx3(config.dbg,
					    n_input, n_hidden, n_output,
					    outlayer_type, 2, 2);
	}
	else
	{
	    QN_ERROR(NULL, "Bunch mode training not yet supported "
		     "in fixed point.");
	    mlp3 = NULL;
	}
// Checking for sysspert() is a hacky way to check to see if
// we are actualy running on a SPERT board.
#ifndef QN_HAVE_SYSSPERT
	QN_WARN(NULL, "fixed point is slow on non-SPERT hosts.");
#endif
#else
	QN_ERROR(NULL, "fixed point not supported in this version.");
	mlp3 = NULL;		// To suppress warning.
#endif
    }
    else
    {
	if (config.bunch_size==1)
	{
	    mlp3 = new QN_MLP_OnlineFl3(config.dbg, "train",
					n_input, n_hidden, n_output,
					outlayer_type);
	}
	else
	{
	    mlp3 = new QN_MLP_BunchFl3(config.dbg, "train",
				       n_input, n_hidden, n_output,
				       outlayer_type, config.bunch_size);
	}
#ifdef QN_HAVE_SYSSPERT
	QN_WARN(NULL, "floating point is very slow on SPERT.");
#endif
    }


    // A weight file of "-" means randomize.
    if (inweights==NULL)
    {
	QN_randomize_weights(config.dbg, config.seed, *mlp3,
			     config.min_random_weight,
			     config.max_random_weight,
			     config.min_random_bias,
			     config.max_random_bias);
	if (verbose>0)
	{
	    QN_OUTPUT("Randomized weights.");
	}
    }
    else
    {
	float min, max;
	QN_MLPWeightFile_RAP3 inwfile(config.dbg, inweights, QN_READ,
				      config.inweights_filename,
				      n_input, n_hidden, n_output);
	QN_read_weights(inwfile, *mlp3, &min, &max, config.dbg);
	QN_OUTPUT("Weights loaded from file, min=%g max=%g.",
		  min, max);
    }

    QN_HardSentTrainer* trainer =
	new QN_HardSentTrainer(config.dbg,          // Debugging level.
			       "trainer",           // Debugging tag.
			       verbose,             // Verbosity level.
			       mlp3,                // MLP.
			       train_winftr_str,    // Training feature strm.
			       train_winlab_str,    // Training label stream.
			       cv_winftr_str,       // CV feature stream.
			       cv_winlab_str,       // CV label stream.
			       rate_sched,	    // Learning rate scheduler.
			       0.0,		    // Low target.
			       1.0,		    // High target.
			       logweights_filename, // Where we log weights.
			       1                    // Batch size.
			       );
    trainer->train();
    delete trainer;
    if (verbose>0)
    {
	QN_OUTPUT("Starting to write weights...");
    }
    float min, max;
    QN_MLPWeightFile_RAP3 outwfile(config.dbg, outweights, QN_WRITE,
				   config.outweights_filename,
				   n_input, n_hidden, n_output);
    QN_write_weights(outwfile, *mlp3, &min, &max, config.dbg);
    if (verbose>0)
    {
	QN_OUTPUT("Weights written.");
    }
// A note for the logfile.
    time(&now);
    if (verbose)
    {
	QN_OUTPUT("Program stop: %.24s", ctime(&now));
    }
    delete mlp3;

    delete cv_winlab_str;
    delete cv_winftr_str;
    delete train_winlab_str;
    delete train_winftr_str;
    if (cv_lab_str!=NULL)
	delete cv_lab_str;
    if (cv_ftr_str!=NULL)
	delete cv_ftr_str;
    if (train_lab_str!=NULL)
	delete train_lab_str;
    if (train_ftr_str!=NULL)
	delete train_ftr_str;
    if (narrow_str!=NULL)
	delete narrow_str;
    if (norm_str!=NULL)
	delete norm_str;
    delete split_lab_str;
    delete split_ftr_str;
    delete ftrlab_str;
    delete rate_sched;
}

int
main(int argc, const char* argv[])
{
    char* progname;		// The name of the prog - set by QN_initargs.
    FILE* db;
    FILE* inweights;
    FILE* outweights;
    FILE* innorms;
    FILE* logfile;
    QN_StreamType db_streamtype; // The file type of the database.
    char logbuf[160];

    set_defaults();
    QN_initargs(&argtab[0], &argc, &argv, &progname);

    // Seed the random number generator.
    // FIXME - this stuff needs sorting out properly
    srand48(config.seed);
    logfile = stdout;
    // *** Warning - setvbuf(?, NULL, _IOLBF, ?) breaks on SPERT.
    if (setvbuf(logfile, logbuf, _IOLBF, sizeof(logbuf)))
	abort();		// Fail on error.
    QN_printargs(logfile, progname, &argtab[0]);
    QN_logger = new QN_Logger_Simple(logfile, stderr, progname);
    
    // Install our own out-of-memory handler if possible.
    // This is important on SPERT, as for some versions of spertdev the
    // default out of memory handler produces no output and is consequently
    // confusing.
#ifdef __GNUC__
    set_new_handler(QN_new_handler);
#endif

    // Check for use of old options.
    if (config.train_cache_sents!=DEFAULT_TRAIN_CACHE_SENTS)
    {
	QN_WARN(NULL, "used old 'train_cache_sents' option - value ignored.");
    }
    if (config.initial_learnrate!=DEFAULT_INITIAL_LEARNRATE)
    {
	QN_WARN(NULL, "used old 'initial_learnrate' option - value ignored.");
    }

    // Some consistency checks.
    if (config.window_width<0 || config.window_width>1000)
    {
	QN_ERROR(NULL, "window must be in range 0-1000.");
    }
    size_t net_num_ftrs;
    if (config.num_ftrs!=0)
	net_num_ftrs = config.num_ftrs;
    else if (config.train_ftrfile_num_ftrs!=0)
	net_num_ftrs = config.train_ftrfile_num_ftrs - config.first_ftr;
    else
	net_num_ftrs = QN_ALL;
    if (net_num_ftrs!=QN_ALL &&
	(size_t) config.window_width*net_num_ftrs != (size_t) config.num_input)
    {
	QN_ERROR(NULL, "number of input features (%i) "
		 "* window size (%i) "
		 "must equal input layer size (%i).",
		 (int) net_num_ftrs, (int) config.window_width,
		 (int) config.num_input);
    }
    if (strcmp(config.train_ftrfile_format, "pfile")==0)
    {
	db_streamtype = QN_STREAM_PFILE;
    }
    else if (strcmp(config.train_ftrfile_format, "pre")==0)
    {
	db_streamtype = QN_STREAM_PREFILE;
    }
    else
    {
	QN_ERROR(NULL, "Unknown file format '%s'.",
		 config.train_ftrfile_format);
	db_streamtype = QN_STREAM_UNKNOWN;
    }

    db = fopen(config.train_ftrfile_filename, "r");
    if (db==NULL)
    {
	QN_ERROR(NULL, "failed to open feature database '%s'.",
		 config.train_ftrfile_filename);
    }
    // Use a bigger buffer for the database file - reduces problems with
    // IO latency on SPERT.
    // *** Warning - setvbuf(?, NULL, _IOFBF, ?) seems to not work on SPERT
    char* buf = new char[0x8000];
    if (setvbuf(db, buf, _IOFBF, 0x8000))
	abort();
    if (strcmp(config.inweights_filename, "-")==0)
	inweights = NULL;
    else
    {
	inweights = fopen(config.inweights_filename, "r");
	if (inweights==NULL)
	{
	    QN_ERROR(NULL, "failed to open input weights file '%s'.",
		     config.inweights_filename);
	}
    }
    if (strcmp(config.innorms_filename, "")!=0)
    {
	innorms = fopen(config.innorms_filename, "r");
	if (innorms==NULL)
	{
	    QN_ERROR(NULL, "failed to open '%s'.", config.innorms_filename);
	}
    }
    else
	innorms = NULL;

    outweights = fopen(config.outweights_filename, "w");
    if (outweights==NULL)
    {
	QN_ERROR(NULL, "failed to open output weights file '%s'.",
		 config.outweights_filename);
    }
    train_run(db, db_streamtype, inweights, innorms, outweights,
	      config.verbose);
    if (inweights!=NULL)
    {
	if (fclose(inweights))
	{
	    QN_ERROR(NULL, "failed to close input weights file %s.",
		     config.inweights_filename);
	}
    }
    if (outweights!=NULL)
    {
	if (fclose(outweights))
	{
	    QN_ERROR(NULL, "failed to close output weights file %s.",
		     config.outweights_filename);
	}
    }
    if (fclose(innorms))
    {
	QN_ERROR(NULL, "failed to close norms file %s.",
		 config.innorms_filename);
    }
    if (fclose(db))
    {
	QN_ERROR(NULL, "failed to close feature database %s.",
		 config.train_ftrfile_filename);
    }
    delete buf;
    exit(EXIT_SUCCESS);
}
