/******************************************************************************
 *
 *	Compute error rate between strings
 *
 *	Inspired and partially dervived from the ICSI wordscore.c package
 *
 *****************************************************************************/

#if !defined(lint) && !defined(SABER)
static char *rcsid = "$Header: /u/drspeech/src/wordscore/RCS/wordscore.cc,v 1.4 1996/10/14 20:09:11 ogawah Exp $ ICSI (Berkeley)";
#endif


#include <stdlib.h>
#include <stdio.h>
#include "sentence.h"
#include "dProg.h"
#include "defines.h"

extern "C" {
extern int getopt(int, char *const *, const char *);
extern char *optarg;
extern int optind;
}

#define TRUE ( 1 == 1 )
#define FALSE ( 1 == 2 )

// global parameters
char *ref_file ;		// reference strings file
char *word_rec_file ;		// recognized strings file
int Verbose ;			// verbose output flag
char *BSIL ;			// begin silence string
char *ESIL ;			// end silence string
char *MSIL ;			// mid-silence string

// local functions
int procCom( int argc, char **argv ) ;
void printUsage( char *name ) ;


//-----------------------------------------------------------------------------
//	Begin wordscore
//-----------------------------------------------------------------------------

main( int argc, char **argv ) {

/*
 *	Initialization
 */
 
  if( !procCom( argc, argv ) ) {
    printUsage( argv[0] ) ;
    exit( 1 ) ;
  }

  // create senetence structures
  sentence *refs = new sentence( ref_file, BSIL, ESIL, MSIL ) ;
  sentence *recs = new sentence(word_rec_file, BSIL, ESIL, MSIL ) ;
  if( refs == NULL || recs == NULL ) {
    perror( "allocating refs and/or recs" ) ;
    exit( 1 ) ;
  }
  
  // create the dynamic programming structure
  dProg *dp = new dProg() ;
  if( dp == NULL ) {
    perror( "allocating dp" ) ;
    exit( 1 ) ;
  }


/*
 *	Score all the  data
 */

  // step through all the recognized utterance
  while( recs->getNext() ) {

    // get the reference utterance
    if( !refs->getNext() ) {
      fprintf( stderr, "%s: fewer refernce utterances than recognized\n", refs->getFileName() ) ;
      exit( 1 ) ;
    }
    if( refs->getNum() == 0 ) {
      fprintf( stderr, "%s: empty reference at line %d\n", refs->getFileName(), refs->getLineNo() );
      exit( 1 ) ;
    }

    // perform the dynamic programming alignment
    dp->align( recs, refs ) ;
    
    // output sentence pairs
    if( Verbose ) {
      dp->printAlign( stdout, recs, refs ) ;
      dp->printSentStats( stdout ) ;
      fprintf( stdout, "\n" ) ;
    }
  }

  // output final statistics
  dp->printStats( stdout ) ;

  exit( 0 );
}

//-----------------------------------------------------------------------------
//	Process the command line options
//-----------------------------------------------------------------------------
int procCom( int argc, char **argv ) {

  // initialize parameters
  ref_file = word_rec_file = NULL ;
  Verbose = FALSE ;
  BSIL = ESIL = MSIL = NULL ;


  // step through the command line
  int c ;
  while( ( c = getopt( argc, argv, "r:w:vb:e:m:") ) != -1 ) {
      
    // find the appropriate setting
    switch( c ) {

    // get the reference file
    case 'r':
      ref_file = optarg ;
      break ;

    // get the recognized file
    case 'w':
      word_rec_file = optarg ;
      break ;
        
    // select verbose output
    case 'v':
      Verbose = TRUE ;
      break ;

    // specify initial silence string
    case 'b':
      BSIL = optarg ;
      break ;

    // specify trailing silence string
    case 'e':
      ESIL = optarg ;
      break ;

    // specify a mid-silence string
    case 'm':
      MSIL = optarg ;
      break ;

    // if you got here, something is wrong
    default:
      fprintf( stderr, "bad command line argument: %c\n", (char) c ) ;
      return 0 ;
    }
  }

  // check for errors
  if( word_rec_file == NULL || ref_file == NULL ) {
    fprintf( stderr, "must specify reference and recognized file\n" ) ;
    return 0 ;
  }

  // check for unspecified
  if( BSIL == NULL ) BSIL = strdup( DEFAULT_BSIL ) ;
  if( ESIL == NULL ) ESIL = strdup( DEFAULT_ESIL ) ;
  if( MSIL == NULL ) MSIL = strdup( DEFAULT_MSIL ) ;

  // indicate OK status
  return 1 ;
}

//-----------------------------------------------------------------------------
//	Print the process usage
//-----------------------------------------------------------------------------
void printUsage( char *name ) {    
  fprintf( stderr, "usage: %s [options] -r <reference-file> ", name ) ;
  fprintf( stderr, "-w <recognized-file>\n" ) ;
  fprintf( stderr, "\t-v                        " ) ;
  fprintf( stderr, "\tverbose output\n" ) ;
  fprintf( stderr, "\t-b <begin-silence-string> " ) ;
  fprintf( stderr, "\tdefault: %s\n", DEFAULT_BSIL ) ;
  fprintf( stderr, "\t-e <end-silence-string>   " ) ;
  fprintf( stderr, "\tdefault: %s\n", DEFAULT_ESIL ) ;
  fprintf( stderr, "\t-m <middle-silence-string>" ) ;
  fprintf( stderr, "\tdefault: %s\n", DEFAULT_MSIL ) ;
}
